from __future__ import annotations

import numpy as np
from abc import ABCMeta, abstractmethod
import wgpu
from .util import load_shader

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .visualizer import Visualizer

class Overlay(metaclass=ABCMeta):

    _blending = {
        "src_factor": wgpu.BlendFactor.src_alpha,
        "dst_factor": wgpu.BlendFactor.one_minus_src_alpha,
        "operation": wgpu.BlendOperation.add
    }

    MAX_INSTANCES = 128
    def __init__(self, visualizer: Visualizer, target_canvas_format=None):
        """Setup the overlay.

        :param visualizer: The visualizer instance
        :param target_texture: The texture to render the overlay to. If None, the visualizer's canvas is used.
        """
        self._visualizer = visualizer
        self._device = self._visualizer.device
        self._contents = None
        if target_canvas_format is None:
            target_canvas_format = self._visualizer.canvas_format
        self._target_canvas_format = target_canvas_format

        # The following are present to allow multiple copies of the same overlay to be
        # rendered. This is used by the periodic_sph module. By default, there is just
        # one copy with weight 1.0.
        self._offset_buffer = self._device.create_buffer(
            label="overlay_offset_buffer",
            size=4*2*self.MAX_INSTANCES,
            usage=wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.COPY_DST
        )
        self._weight_buffer = self._device.create_buffer(
            label="overlay_weight_buffer",
            size=4 * self.MAX_INSTANCES,
            usage=wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.COPY_DST
        )

        self._setup_shader_module()
        self._setup_texture()
        self._setup_sampler()
        self._setup_params_buffer()
        self._setup_render_pipeline()

    def _setup_shader_module(self):
        if not hasattr(self._visualizer, "_overlay_shader"):
            # this looks a bit weird, but it's because we want to share the shader module
            # between all overlays using the same device
            self._visualizer._overlay_shader = self._device.create_shader_module(code=load_shader("overlay.wgsl"),
                                                                                 label="overlay")

    @property
    def _shader(self):
        return self._visualizer._overlay_shader

    def _setup_sampler(self):
        self._sampler = self._device.create_sampler(label="overlay_sampler",
                                                    mag_filter=wgpu.FilterMode.linear,
                                                    min_filter=wgpu.FilterMode.linear)

    def _setup_texture(self):
        im = self.get_contents()

        assert len(im.shape)==3 and im.shape[2]==4, "Overlay must be RGBA image"
        assert im.dtype==np.float32, "Overlay must be float32"

        self._texture = self._device.create_texture(
            label="overlay_texture",
            size=(im.shape[1], im.shape[0], 1),
            dimension=wgpu.TextureDimension.d2,
            format=wgpu.TextureFormat.rgba32float,
            mip_level_count=1,
            sample_count=1,
            usage=wgpu.TextureUsage.COPY_DST | wgpu.TextureUsage.TEXTURE_BINDING)

        self._device.queue.write_texture(
            {
                "texture": self._texture,
                "mip_level": 0,
                "origin": [0, 0, 0],
            },
            im.tobytes(),
            {
                "bytes_per_row": 4 * 4 * im.shape[1],
                "offset": 0,
            },
            (im.shape[1], im.shape[0], 1)
        )



    def _setup_params_buffer(self):
        self._overlay_params_buffer = self._device.create_buffer(
            label="overlay_params_buffer",
            size=4*8,
            usage=wgpu.BufferUsage.UNIFORM | wgpu.BufferUsage.COPY_DST
        )

    def get_texturespace_coordinates(self, width, height) -> tuple[float,float,float,float]:
        return (0.0,0.0,1.0,1.0)

    def _update_params_buffer(self, width, height):
        x, y, w, h = self.get_clipspace_coordinates(width, height)
        x_t, y_t, w_t, h_t = self.get_texturespace_coordinates(width, height)
        self._device.queue.write_buffer(self._overlay_params_buffer, 0,
                                        np.array([x,y,w,h,x_t,y_t,w_t,h_t], dtype=np.float32).tobytes())
    def _setup_render_pipeline(self):
        self._bind_group_layout = self._device.create_bind_group_layout(
            label="overlay_bind_group_layout",
            entries=[
                {
                    "binding": 0,
                    "visibility": wgpu.ShaderStage.VERTEX,
                    "buffer": {
                        "type": wgpu.BufferBindingType.uniform,
                    },
                },
                {
                    "binding": 1,
                    "visibility": wgpu.ShaderStage.FRAGMENT,
                    "texture": {
                        "sample_type": wgpu.TextureSampleType.float,
                        "view_dimension": wgpu.TextureViewDimension.d2
                    },
                },
                {
                    "binding": 2,
                    "visibility": wgpu.ShaderStage.FRAGMENT,
                    "sampler": {
                        "type": wgpu.SamplerBindingType.filtering,
                    },
                },
            ],
        )

        self._bind_group = self._device.create_bind_group(
            label="overlay_bind_group",
            layout=self._bind_group_layout,
            entries=[
                {
                    "binding": 0,
                    "resource": {
                        "buffer": self._overlay_params_buffer,
                        "offset": 0,
                        "size": 4*8,
                    },
                },
                {
                    "binding": 1,
                    "resource": self._texture.create_view(),
                },
                {
                    "binding": 2,
                    "resource": self._sampler,
                },
            ],
        )

        self._pipeline_layout = self._device.create_pipeline_layout(
            label="overlay_render_pipeline_layout",
            bind_group_layouts=[self._bind_group_layout]
        )

        self._render_pipeline = self._device.create_render_pipeline(
            label="overlay_render_pipeline",
            layout=self._pipeline_layout,
            vertex={
                "module": self._shader,
                "entry_point": "vertex_main",
                "buffers": [{
                            "array_stride": 8,
                            "step_mode": wgpu.VertexStepMode.instance,
                            "attributes": [
                                {
                                    "format": wgpu.VertexFormat.float32x2,
                                    "offset": 0,
                                    "shader_location": 0,
                                }
                            ]
                           },
                            {
                                "array_stride": 4,
                                "step_mode": wgpu.VertexStepMode.instance,
                                "attributes": [
                                    {
                                        "format": wgpu.VertexFormat.float32,
                                        "offset": 0,
                                        "shader_location": 1,
                                    }
                                ]
                            }
                ]
            },
            primitive={
                "topology": wgpu.PrimitiveTopology.triangle_strip,
            },
            depth_stencil=None,
            multisample=None,
            fragment={
                "module": self._shader,
                "entry_point": "fragment_main",
                "targets": [
                    {
                        "format": self._target_canvas_format,
                        "blend": {
                            "color": self._blending,
                            "alpha": self._blending,
                        },
                        "write_mask": wgpu.ColorWrite.ALL,
                    }
                ]
            }
        )

    def get_instance_offsets_and_weights(self):
        return np.array([[0.0,0.0]], dtype=np.float32), np.ones(1, dtype=np.float32)

    def encode_render_pass(self, command_encoder: wgpu.GPUCommandEncoder,
                           target_texture_view: wgpu.GPUTextureView,
                           clear=False):
        self._update_params_buffer(target_texture_view.size[0], target_texture_view.size[1])

        render_pass = command_encoder.begin_render_pass(
            color_attachments=[{
                "view": target_texture_view,
                "resolve_target": None,
                "clear_value": (0.0, 0.0, 0.0, 1.0),
                "load_op": wgpu.LoadOp.clear if clear else wgpu.LoadOp.load,
                "store_op": wgpu.StoreOp.store,
            }],
        )

        instance_offsets, instance_weights = self.get_instance_offsets_and_weights()
        assert len(instance_offsets)<=self._offset_buffer.size//4//2, "Too many instances for offset buffer"
        self._device.queue.write_buffer(self._offset_buffer, 0, instance_offsets.tobytes())
        self._device.queue.write_buffer(self._weight_buffer, 0, instance_weights.tobytes())

        render_pass.set_pipeline(self._render_pipeline)
        render_pass.set_vertex_buffer(0, self._offset_buffer)
        render_pass.set_vertex_buffer(1, self._weight_buffer)
        render_pass.set_bind_group(0, self._bind_group, [], 0, 99999)
        render_pass.draw(4, len(instance_offsets), 0, 0)
        render_pass.end()

    @abstractmethod
    def get_clipspace_coordinates(self, width, height) -> tuple[float,float,float,float]:
        """Must return geometry of overlay as x0, y0, width, height in clip space coordinates."""
        pass

    def get_contents(self) -> np.ndarray:
        """Return a cached 2D image with RGBA channels for display.

        If needed, invokes a render_contents() call to the subclass."""
        if self._contents is None:
            self._contents = self.render_contents()
        return self._contents

    def invalidate_contents(self):
        """Mark the contents as invalid, needing a re-render"""
        self._contents = None
    @abstractmethod
    def render_contents(self) -> np.ndarray:
        """Must return a 2D image with RGBA channels for display."""
        pass

    def update(self):
        self._contents = None
        self._setup_texture()
        self._setup_render_pipeline()


from __future__ import annotations

import logging
import numpy as np
import time
import wgpu

from contextlib import contextmanager
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

from . import config
from . import canvas
from . import colormap
from . import sph, periodic_sph
from . import colorbar
from . import text
from . import scalebar
from . import loader
from . import util
from . import line
from . import simcube
from . import view_synchronizer
from . import particle_buffers
from .drawreason import DrawReason

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class VisualizerBase:
    colorbar_aspect_ratio = config.COLORBAR_ASPECT_RATIO
    show_status = True
    device = None # device will be shared across all instances

    def __init__(self, data_loader_class = loader.TestDataLoader, data_loader_args = (), data_loader_kwargs={},
                 *, render_resolution = config.DEFAULT_RESOLUTION, periodic_tiling = False,
                 colormap_name = config.DEFAULT_COLORMAP, canvas_class = canvas.VisualizerCanvas,
                 hdr = False, rgb=False):
        self._hdr = hdr
        self._rgb = rgb
        self._colormap_name = colormap_name
        self._render_resolution = render_resolution
        self._sph_class = sph.SPH
        self._colorbar: Optional[colorbar.ColorbarOverlay] = None
        self._colormap: Optional[colormap.Colormap] = None

        self._encoder_executor = ThreadPoolExecutor(max_workers=1) # 1 worker to prevent GIL contention

        self.crosshairs_visible = False

        self._prevent_sph_rendering = False # when True, prevents the sph from rendering, to ensure quick screen updates

        self.show_colorbar = True
        self.show_scalebar = True

        self.canvas = canvas_class(visualizer=self, title="topsy")

        self._setup_wgpu()

        self.data_loader = data_loader_class(self.device, *data_loader_args, **data_loader_kwargs)
        self.particle_buffers = particle_buffers.ParticleBuffers(self.data_loader, self.device,
                                                                 self.data_loader.get_render_progression().get_max_particle_regions_per_block())

        self.periodicity_scale = self.data_loader.get_periodicity_scale()

        self._periodic_tiling = periodic_tiling

        if periodic_tiling:
            self._sph = periodic_sph.PeriodicSPH(self, self._render_resolution)
        elif self._rgb:
            logger.info("Using RGB renderer")
            self._sph = sph.RGBSPH(self, self._render_resolution)
        else:
            self._sph = sph.SPH(self, self._render_resolution)

        self.render_texture = self._sph.get_output_texture()

        self.reset_view()

        self._reinitialize_colormap_and_bar()

        self._last_status_update = 0.0
        self._status = text.TextOverlay(self, "topsy", (-0.9, 0.9), 40, color=(1, 1, 1, 1))

        self._scalebar = scalebar.ScalebarOverlay(self)

        self._crosshairs = line.Line(self,
                                     [(-1, 0,0,0), (1, 0,0,0),
                                      (200,200,0,0),
                                      (0, 1, 0, 0), (0, -1, 0, 0)],
                                     (1, 1, 1, 0.3) # color
                                     , 10.0)
        self._cube = simcube.SimCube(self, (1, 1, 1, 0.3), 10.0)

    def _setup_wgpu(self):
        self.adapter: wgpu.GPUAdapter = wgpu.gpu.request_adapter_sync(power_preference="high-performance")
        logger.info(f"GPU device: {self.adapter.info['adapter_type']} ({self.adapter.info['device']})")
        if self.device is None:
            max_buffer_size = self.adapter.limits['max-buffer-size']
            # on some systems, this is 2^64 which can lead to overflows
            if max_buffer_size > 2**63:
                max_buffer_size = 2**63
            type(self).device: wgpu.GPUDevice = self.adapter.request_device_sync(
                required_features=["texture-adapter-specific-format-features", "float32-filterable",
                                  "multi-draw-indirect"],
                required_limits={"max_buffer_size": max_buffer_size})
        self.context: wgpu.GPUCanvasContext = self.canvas.get_context("wgpu")

        if self._hdr:
            self.canvas_format = "rgba16float"
        else:
            self.canvas_format = self.context.get_preferred_format(self.adapter)
            if self.canvas_format.endswith("-srgb"):
                # matplotlib colours aren't srgb. It might be better to convert
                # but for now, just stop the canvas being srgb
                self.canvas_format = self.canvas_format[:-5]

        self.context.configure(device=self.device, format=self.canvas_format)

        logger.info(f"Canvas format {self.canvas_format}")

    def invalidate(self, reason=DrawReason.CHANGE):
        # NB no need to check if we're already pending a draw - rendercanvas does that for us
        self.canvas.request_draw(lambda: self.draw(reason))

    def rotate(self, x_angle, y_angle):
        dx_rotation_matrix = self._x_rotation_matrix(x_angle)
        dy_rotation_matrix = self._y_rotation_matrix(y_angle)
        self.rotation_matrix = dx_rotation_matrix @ dy_rotation_matrix @ self.rotation_matrix

    @property
    def rotation_matrix(self):
        return self._sph.rotation_matrix

    @rotation_matrix.setter
    def rotation_matrix(self, value):
        self._sph.rotation_matrix = value
        self.invalidate()

    @property
    def colormap_name(self):
        return self._colormap_name

    @property
    def colormap(self):
        return self._colormap

    @colormap_name.setter
    def colormap_name(self, value):
        self._colormap_name = value
        self._reinitialize_colormap_and_bar(keep_scale=True)
        self.invalidate(reason=DrawReason.PRESENTATION_CHANGE)

    @property
    def position_offset(self):
        return self._sph.position_offset

    @position_offset.setter
    def position_offset(self, value):
        self._sph.position_offset = value
        self.invalidate()

    def reset_view(self):
        self._sph.rotation_matrix = np.eye(3)
        period_scale = self.data_loader.get_periodicity_scale()
        if period_scale is not None:
            self.scale = period_scale / 2
        else:
            self.scale = config.DEFAULT_SCALE
        self._sph.position_offset = np.zeros(3)

    @property
    def scale(self):
        """Return the scalefactor from kpc to viewport coordinates. Viewport will therefore be 2*scale wide."""
        return self._sph.scale
    
    @scale.setter
    def scale(self, value):
        self._sph.scale = value
        self.invalidate()

    @property
    def quantity_name(self):
        """The name of the quantity being visualised, or None if density projection."""
        return self.particle_buffers.quantity_name

    @property
    def averaging(self):
        """True if the quantity being visualised is a weighted average, False if it is a mass projection."""
        return self.particle_buffers.quantity_name is not None

    @quantity_name.setter
    def quantity_name(self, value):
        if value == self.particle_buffers.quantity_name:
            return

        if value is not None:
            # see if we can get it. Assume it'll be cached, so this won't waste time.
            try:
                self.data_loader.get_named_quantity(value)
            except Exception as e:
                raise ValueError(f"Unable to get quantity named '{value}'") from e

        self.particle_buffers.quantity_name = value
        
        self._reinitialize_colormap_and_bar()
        

    def _reinitialize_colormap_and_bar(self, keep_scale=False):
        """Reinitialize the colormap and colorbar.
        
        If keep_scale is False, render and figure out the min/max values for the colormap too.
        """
        if self._colormap is not None:
            old_vmin, old_vmax, old_log = self._colormap.vmin, self._colormap.vmax, self._colormap.log_scale
        else:
            if keep_scale:
                raise ValueError("Cannot keep scale if colormap not yet initialized")

        if self._rgb:
            if self._hdr:
                self._colormap = colormap.RGBHDRColormap(self)
            else:
                self._colormap = colormap.RGBColormap(self)
        elif self._hdr:
            self._colormap = colormap.HDRColormap(self, weighted_average=self.quantity_name is not None)

        else:
            self._colormap = colormap.Colormap(self, weighted_average=self.quantity_name is not None)
            
        if not keep_scale:
            self._sph.render(DrawReason.EXPORT)
            self._colormap.autorange_vmin_vmax()
        else:
            self._colormap.vmin = old_vmin
            self._colormap.vmax = old_vmax
            self._colormap.log_scale = old_log

        if not self._rgb and not self._hdr:
            self._colorbar = colorbar.ColorbarOverlay(self, self.vmin, self.vmax, self.colormap_name, self._get_colorbar_label())

        self.invalidate(DrawReason.PRESENTATION_CHANGE)


    def _get_colorbar_label(self):
        label = self.data_loader.get_quantity_label(self.quantity_name)
        if self._colormap.log_scale:
            label = r"$\log_{10}$ " + label
        return label

    @staticmethod
    def _y_rotation_matrix(angle):
        return np.array([[1, 0, 0],
                         [0, np.cos(angle), -np.sin(angle)],
                         [0, np.sin(angle), np.cos(angle)]])

    @staticmethod
    def _x_rotation_matrix(angle):
        return np.array([[np.cos(angle), 0, np.sin(angle)],
                         [0, 1, 0],
                         [-np.sin(angle), 0, np.cos(angle)]])

    @contextmanager
    def prevent_sph_rendering(self):
        self._prevent_sph_rendering = True
        try:
            yield
        finally:
            self._prevent_sph_rendering = False

    def _encode_draw(self, target_texture_view):
        command_encoder = self.device.create_command_encoder()

        self._colormap.encode_render_pass(command_encoder, target_texture_view)
        if self.show_colorbar and self._colorbar is not None:
            self._colorbar.encode_render_pass(command_encoder, target_texture_view)
        if self.show_scalebar:
            self._scalebar.encode_render_pass(command_encoder, target_texture_view)
        if self.crosshairs_visible:
            self._crosshairs.encode_render_pass(command_encoder, target_texture_view)
        if self._periodic_tiling:
            self._cube.encode_render_pass(command_encoder, target_texture_view)

        if self.show_status:
            self._update_and_display_status(command_encoder, target_texture_view)

        result = command_encoder.finish()
        return result

    def draw(self, reason, target_texture_view=None):

        if target_texture_view is None:
            target_texture_view = self.canvas.get_context("wgpu").get_current_texture().create_view()

        command_buffer_future = self._encoder_executor.submit(self._encode_draw, target_texture_view)

        if not self._prevent_sph_rendering:
            self.render_sph(reason)

        self._colormap.set_scaling(target_texture_view,
                                   self._sph.last_render_mass_scale if not self.averaging else 1.0)

        self.device.queue.submit([command_buffer_future.result()])

        if reason != DrawReason.EXPORT and (not self._prevent_sph_rendering):
            if self._sph.needs_refine():
                self.invalidate(DrawReason.REFINE)

    def render_sph(self, draw_reason = DrawReason.CHANGE):
        self._sph.render(draw_reason)

    @property
    def vmin(self):
        return self._colormap.vmin

    @property
    def vmax(self):
        return self._colormap.vmax

    @vmin.setter
    def vmin(self, value):
        self._colormap.vmin = value
        self._refresh_colorbar()
        self.invalidate(DrawReason.PRESENTATION_CHANGE)

    @vmax.setter
    def vmax(self, value):
        self._colormap.vmax = value
        self._refresh_colorbar()
        self.invalidate(DrawReason.PRESENTATION_CHANGE)

    @property
    def gamma(self):
        if hasattr(self._colormap, 'gamma'):
            return self._colormap.gamma
        else:
            return None

    @gamma.setter
    def gamma(self, value):
        if hasattr(self._colormap, 'gamma'):
            self._colormap.gamma = value
            self.invalidate(DrawReason.PRESENTATION_CHANGE)

    @property
    def log_scale(self):
        return self._colormap.log_scale

    @log_scale.setter
    def log_scale(self, value):
        self._colormap.log_scale = value
        self._refresh_colorbar()
        self.invalidate(DrawReason.PRESENTATION_CHANGE)

    def _refresh_colorbar(self):
        if self._colorbar is None:
            return
        self._colorbar.vmin = self._colormap.vmin
        self._colorbar.vmax = self._colormap.vmax
        self._colorbar.label = self._get_colorbar_label()
        self._colorbar.update()

    def sph_clipspace_to_screen_clipspace_matrix(self):
        aspect_ratio = self.canvas.width_physical / self.canvas.height_physical
        if aspect_ratio>1:
            y_squash = aspect_ratio
            x_squash = 1.0
        elif aspect_ratio<1:
            y_squash = 1.0
            x_squash = 1.0/aspect_ratio
        else:
            x_squash = 1.0
            y_squash = 1.0

        matr = np.eye(4, dtype=np.float32)
        matr[0,0] = x_squash
        matr[1,1] = y_squash
        return matr



    def display_status(self, text, timeout=0.5):
        self._override_status_text = text
        self._override_status_text_until = time.time()+timeout

    def _update_and_display_status(self, command_encoder, target_texture_view):
        now = time.time()
        if hasattr(self, "_override_status_text_until") and now<self._override_status_text_until:
            if self._status.text!=self._override_status_text and now-self._last_status_update>config.STATUS_LINE_UPDATE_INTERVAL_RAPID:
                self._status.text = self._override_status_text
                self._last_status_update = now
                self._status.update()

        elif now - self._last_status_update > config.STATUS_LINE_UPDATE_INTERVAL:
            self._last_status_update = now
            self._status.text = f"${self._sph.last_render_fps:.0f}$ fps"
            factor = np.round(self._sph.last_render_mass_scale, 1)
            if factor>1.1:
                self._status.text += f" /{factor:.1f}ds"
            geom_factor = self._sph._render_progression.get_fraction_volume_selected()
            if geom_factor<0.9:
                self._status.text += f" /{1./geom_factor:.1f}gf"

            self._status.update()

        self._status.encode_render_pass(command_encoder, target_texture_view)

    def get_sph_image(self) -> np.ndarray:
        return self._sph.get_image(self.averaging)

    def get_depth_image(self) -> np.ndarray:
        return self._sph.get_depth_image()


    def get_presentation_image(self, resolution=(640,480)) -> np.ndarray:
        texture: wgpu.GPUTexture = self.device.create_texture(
            size=(resolution[0], resolution[1], 1),
            usage=wgpu.TextureUsage.RENDER_ATTACHMENT |
                  wgpu.TextureUsage.COPY_SRC,
            format=self.canvas_format,
            label="output_texture",
        )
        self.draw(DrawReason.EXPORT, texture.create_view())

        size = texture.size

        if texture.format.endswith("8unorm"):
            bytes_per_pixel = 4
            np_type = np.uint8
        elif texture.format.endswith("16float"):
            bytes_per_pixel = 8
            np_type = np.float16
        else:
            raise ValueError(f"Unsupported texture format {texture.format}")

        data = self.device.queue.read_texture(
            {
                "texture": texture,
                "origin": (0, 0, 0),
            },
            {
                "offset": 0,
                "bytes_per_row": bytes_per_pixel * size[0],
                "rows_per_image": size[1],
            },
            size,
        )

        return np.frombuffer(data, np_type).reshape(size[1], size[0], 4)


    def save(self, filename='output.pdf'):
        image = self.get_sph_image()
        import matplotlib.pyplot as p
        fig = p.figure()
        p.clf()
        p.set_cmap(self.colormap_name)
        extent = np.array([-1., 1., -1., 1.])*self.scale
        if self._colormap.log_scale:
            image = np.log10(image)

        p.imshow(image,
                 vmin=self._colormap.vmin,
                 vmax=self._colormap.vmax,
                 extent=extent)
        p.xlabel("$x$/kpc")
        p.colorbar().set_label(self._colorbar.label)
        p.savefig(filename)
        p.close(fig)

    def show(self, force=False):
        from rendercanvas import jupyter
        if isinstance(self.canvas, jupyter.RenderCanvas):
            return self.canvas
        else:
            from rendercanvas import qt # can only safely import this if we think we're running in a qt environment
            assert isinstance(self.canvas, qt.RenderCanvas)
            self.canvas.show()
            if force or not util.is_inside_ipython():
                qt.loop.run()
            elif not util.is_ipython_running_qt_event_loop():
                # is_inside_ipython_console must be True; if it were False, the previous branch would have run
                # instead.
                print("\r\nYou appear to be running from inside ipython, but the gui event loop is not running.\r\n"
                      "Please run %gui qt in ipython before calling show().\r\n"
                      "\r\n"
                      "Alternatively, if you do not want to continue interacting with ipython while the\r\n"
                      "visualizer is running, you can call show(force=True) to run the gui without access\r\n"
                      "to the ipython console until you close the visualizer window.\r\n\r\n"
                      )
        #else:
        #    raise RuntimeError("The wgpu library is using a gui backend that topsy does not recognize")

    def _ipython_display_(self):
        if isinstance(self.canvas, canvas.jupyter.VisualizerCanvas):
            self.canvas.ipython_display_with_widgets()
        else:
            from IPython.display import display
            display(repr(self))

class Visualizer(view_synchronizer.SynchronizationMixin, VisualizerBase):
    pass


from __future__ import annotations


import PySide6 # noqa: F401 (need to import to select the qt backend)
from PySide6 import QtWidgets, QtGui, QtCore

from rendercanvas.qt import RenderCanvas, loop

from .colormap import RGBColorControls, ColorMapControls
from .lineedit import MyLineEdit
from .recording import RecordingSettingsDialog, VisualizationRecorderWithQtProgressbar
from .. import VisualizerCanvasBase
from ...drawreason import DrawReason

import os
import logging

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    pass


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

def _get_icon(name):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    return QtGui.QIcon(os.path.join(this_dir, "icons", name))


class VisualizerCanvas(VisualizerCanvasBase, RenderCanvas):

    _all_instances = []
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._all_instances.append(self)
        self.hide()

        self._toolbar = QtWidgets.QToolBar()
        self._toolbar.setIconSize(QtCore.QSize(16, 16))

        # setup toolbar to show text and icons
        self._toolbar.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonTextBesideIcon)

        self._load_icons()

        self._record_action = QtGui.QAction(self._record_icon, "Record", self)
        self._record_action.triggered.connect(self.on_click_record)

        self._save_action = QtGui.QAction(self._save_icon, "Snapshot", self)
        self._save_action.triggered.connect(self.on_click_save)

        self._save_movie_action = QtGui.QAction(self._save_movie_icon, "Save mp4", self)
        self._save_movie_action.triggered.connect(self.on_click_save_movie)
        self._save_movie_action.setDisabled(True)

        self._save_script_action = QtGui.QAction(self._export_icon, "Save timestream", self)
        self._save_script_action.triggered.connect(self.on_click_save_script)
        self._save_script_action.setDisabled(True)

        self._load_script_action = QtGui.QAction(self._import_icon, "Load timestream", self)
        self._load_script_action.triggered.connect(self.on_click_load_script)

        self._link_action = QtGui.QAction(self._unlinked_icon, "Link to other windows", self)
        self._link_action.setIconText("Link")
        self._link_action.triggered.connect(self.on_click_link)




        self._cmap_icon = _get_icon("rgb.png")
        self._open_cmap = QtGui.QAction(self._cmap_icon, "Color", self)



        self._toolbar.addAction(self._load_script_action)
        self._toolbar.addAction(self._save_script_action)
        self._toolbar.addAction(self._record_action)
        self._toolbar.addAction(self._save_movie_action)

        self._toolbar.addSeparator()
        self._toolbar.addAction(self._save_action)
        self._toolbar.addSeparator()

        self._toolbar.addAction(self._open_cmap)
        self._toolbar.addSeparator()


        self._toolbar.addAction(self._link_action)
        self._recorder = None



        # now replace the wgpu layout with our own
        layout = self.layout()
        layout.removeWidget(self._subwidget)

        our_layout = PySide6.QtWidgets.QVBoxLayout()
        our_layout.addWidget(self._subwidget)
        our_layout.addWidget(self._toolbar)
        our_layout.setContentsMargins(0, 0, 0, 0)
        our_layout.setSpacing(0)

        self._toolbar.adjustSize()

        self._toolbar_update_timer = QtCore.QTimer(self)
        self._toolbar_update_timer.timeout.connect(self._update_toolbar)
        self._toolbar_update_timer.start(100)

        layout.addLayout(our_layout)
        self.call_later(0, self._on_ready_callback)

    def _on_ready_callback(self):
        if not self._visualizer._hdr and not self._visualizer._rgb:
            self._colormap_controls = ColorMapControls(self)
        elif self._visualizer._rgb:
            self._colormap_controls = RGBColorControls(self)
        self._open_cmap.triggered.connect(self._colormap_controls.open)


    def __del__(self):
        try:
            self._all_instances.remove(self)
        except ValueError:
            pass
        super().__del__()

    def _load_icons(self):
        self._record_icon = _get_icon("record.png")
        self._stop_icon = _get_icon("stop.png")
        self._save_icon = _get_icon("camera.png")
        self._linked_icon = _get_icon("linked.png")
        self._unlinked_icon = _get_icon("unlinked.png")
        self._save_movie_icon = _get_icon("movie.png")
        self._import_icon = _get_icon("load_script.png")
        self._export_icon = _get_icon("save_script.png")

    def on_click_record(self):

        if self._recorder is None or not self._recorder.recording:
            logger.info("Starting recorder")
            self._recorder = VisualizationRecorderWithQtProgressbar(self._visualizer, self)
            self._recorder.record()
            self._record_action.setIconText("Stop")
            self._record_action.setIcon(self._stop_icon)
        else:
            logger.info("Stopping recorder")
            self._recorder.stop()
            self._record_action.setIconText("Record")
            self._record_action.setIcon(self._record_icon)

    def on_click_save_movie(self):
        # show the options dialog first:
        dialog = RecordingSettingsDialog(self)
        dialog.exec()
        if dialog.result() == QtWidgets.QDialog.DialogCode.Accepted:
            fd = QtWidgets.QFileDialog(self)
            fname, _ = fd.getSaveFileName(self, "Save video", "", "MP4 (*.mp4)")
            if fname:
                logger.info("Saving video to %s", fname)
                self._recorder.save_mp4(fname, show_colorbar=dialog.show_colorbar,
                                        show_scalebar=dialog.show_scalebar,
                                        fps=dialog.fps,
                                        resolution=dialog.resolution,
                                        smooth=dialog.smooth,
                                        set_vmin_vmax=dialog.set_vmin_vmax,
                                        set_quantity=dialog.set_quantity)
                QtGui.QDesktopServices.openUrl(QtCore.QUrl.fromLocalFile(fname))

    def on_click_save(self):
        fd = QtWidgets.QFileDialog(self)
        fname, _ = fd.getSaveFileName(self, "Save snapshot", "", "PNG (*.png);; PDF (*.pdf)")
        if fname:
            logger.info("Saving snapshot to %s", fname)
            self._visualizer.save(fname)
            QtGui.QDesktopServices.openUrl(QtCore.QUrl.fromLocalFile(fname))

    def on_click_save_script(self):
        fd = QtWidgets.QFileDialog(self)
        fname, _ = fd.getSaveFileName(self, "Save camera movements", "", "Python Pickle (*.pickle)")
        if fname:
            logger.info("Saving timestream to %s", fname)
            self._recorder.save_timestream(fname)

    def on_click_load_script(self):
        fd = QtWidgets.QFileDialog(self)
        fname, _ = fd.getOpenFileName(self, "Load camera movements", "", "Python Pickle (*.pickle)")
        if fname:
            logger.info("Loading timestream from %s", fname)
            self._recorder = VisualizationRecorderWithQtProgressbar(self._visualizer, self)
            self._recorder.load_timestream(fname)


    def on_click_link(self):
        if self._visualizer.is_synchronizing():
            logger.info("Stop synchronizing")
            self._visualizer.stop_synchronizing()
        else:
            logger.info("Start synchronizing")
            from ... import view_synchronizer
            synchronizer = view_synchronizer.ViewSynchronizer()
            for instance in self._all_instances:
                synchronizer.add_view(instance._visualizer)

    def _update_toolbar(self):
        if self._recorder is not None or len(self._all_instances)<2:
            self._link_action.setDisabled(True)
        else:
            self._link_action.setDisabled(False)
            if self._visualizer.is_synchronizing():
                self._link_action.setIcon(self._linked_icon)
                self._link_action.setIconText("Unlink")
            else:
                self._link_action.setIcon(self._unlinked_icon)
                self._link_action.setIconText("Link")
        if self._recorder is not None and not self._recorder.recording:
            self._save_movie_action.setDisabled(False)
            self._save_script_action.setDisabled(False)
        else:
            self._save_movie_action.setDisabled(True)
            self._save_script_action.setDisabled(True)




    def request_draw(self, function=None):
        # As a side effect, wgpu gui layer stores our function call, to enable it to be
        # repainted later. But we want to distinguish such repaints and handle them
        # differently, so we need to replace the function with our own
        call_count = 0
        def function_wrapper():
            nonlocal call_count
            if call_count == 0:
                function()
            else:
                # we have been cached!
                self._visualizer.draw(DrawReason.PRESENTATION_CHANGE)
            call_count += 1

        super().request_draw(function_wrapper)

    @classmethod
    def call_later(cls, delay, fn, *args):
        loop.call_later(delay, fn, *args)



from __future__ import annotations

import numpy as np
import wgpu

from .util import load_shader

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .visualizer import Visualizer
class Line:
    def __init__(self, visualizer: Visualizer, path, color, width):
        self._visualizer = visualizer

        if path is not None:
            path = np.asarray(path, dtype=np.float32)
            assert path.ndim == 2, "Path must be an array of points, each with 4 (xyzw) coordinates"
            assert path.shape[1] == 4, "Path must be an array of points, each with 4 (xyzw) coordinates"
            self._line_starts = path[:-1]
            self._line_ends = path[1:]
        else:
            assert hasattr(self, "_line_starts") and hasattr(self, "_line_ends"), \
                "Either path must be provided, or _line_starts and _line_ends must be defined by a subclass"
            assert len(self._line_starts) == len(self._line_ends), \
                "Number of line starts must equal number of line ends"

        self._color = color
        self._width = width
        self._device = visualizer.device
        self._target_canvas_format = visualizer.canvas_format

        self._setup_shader_module()
        self._setup_buffers()
        self._setup_render_pipeline()

    def _setup_shader_module(self):
        self._shader_module = self._visualizer.device.create_shader_module(
            code=load_shader("line.wgsl"),
            label="line_shader_module"
        )


    def _setup_buffers(self):
        self._vertex_buffer_starts = self._visualizer.device.create_buffer_with_data(
            usage=wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.COPY_DST,
            label="line_vertex_buffer_start",
            data=self._line_starts
        )

        self._vertex_buffer_ends = self._visualizer.device.create_buffer_with_data(
            usage=wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.COPY_DST,
            label="line_vertex_buffer_start",
            data=self._line_ends
        )

        _param_dtype = np.dtype([
            ("transform", np.float32, (4, 4)),
            ("color", np.float32, 4),
            ("vp_size_pix", np.float32, 2),
            ("width_pix", np.float32),
            ("padding", np.float32, 3)
        ])

        self._param_buffer = self._device.create_buffer(
            label="line_param_buffer",
            size=_param_dtype.itemsize,
            usage=wgpu.BufferUsage.UNIFORM | wgpu.BufferUsage.COPY_DST
        )

        self._params = np.zeros(1, dtype=_param_dtype)
        self._params["transform"] = np.eye(4)
        self._params["color"] = self._color
        self._params["width_pix"] = self._width

    def _setup_render_pipeline(self):
        self._bind_group_layout = self._device.create_bind_group_layout(
            label="line_bind_group_layout",
            entries=[
                {
                    "binding": 0,
                    "visibility": wgpu.ShaderStage.VERTEX,
                    "buffer": {
                        "type": wgpu.BufferBindingType.uniform,
                    }
                }
            ]
        )

        self._bind_group = self._device.create_bind_group(
            label="line_bind_group",
            layout=self._bind_group_layout,
            entries=[
                {
                    "binding": 0,
                    "resource": {
                        "buffer": self._param_buffer,
                        "offset": 0,
                        "size": self._param_buffer.size
                    }
                }
            ]
        )

        self._pipeline_layout = self._device.create_pipeline_layout(
            label="line_pipeline_layout",
            bind_group_layouts=[self._bind_group_layout]
        )

        self._render_pipeline = self._device.create_render_pipeline(
            label="line_render_pipeline",
            layout=self._pipeline_layout,
            vertex={
                "module": self._shader_module,
                "entry_point": "vertex_main",
                "buffers": [
                    {   # start of line segment
                        "array_stride": 4*4,
                        "step_mode": wgpu.VertexStepMode.instance,
                        "attributes": [
                            {
                                "format": wgpu.VertexFormat.float32x4,
                                "offset": 0,
                                "shader_location": 0
                            }
                        ]
                    },
                    {   # end of line segment
                        "array_stride": 4 * 4,
                        "step_mode": wgpu.VertexStepMode.instance,
                        "attributes": [
                            {
                                "format": wgpu.VertexFormat.float32x4,
                                "offset": 0,
                                "shader_location": 1
                            }
                        ]
                    }
                ]
            },
            primitive={
                "topology": wgpu.PrimitiveTopology.triangle_strip,
            },
            depth_stencil=None,
            multisample=None,
            fragment={
                "module": self._shader_module,
                "entry_point": "fragment_main",
                "targets": [
                    {
                        "format": self._target_canvas_format,
                        "blend": {
                            "color": {
                                 "src_target": wgpu.BlendFactor.src_alpha,
                                 "dst_target": wgpu.BlendFactor.one_minus_src_alpha,
                                 "operation": wgpu.BlendOperation.add,
                            },
                            "alpha": {
                                "src_target": wgpu.BlendFactor.src_alpha,
                                "dst_target": wgpu.BlendFactor.one_minus_src_alpha,
                                "operation": wgpu.BlendOperation.add,
                            },
                        "write_mask": wgpu.ColorWrite.ALL,
                        }
                    }
                ]
            }
        )


    def encode_render_pass(self, command_encoder: wgpu.GPUCommandEncoder,
                           target_texture_view: wgpu.GPUTextureView):

        self._params["vp_size_pix"] = target_texture_view.size[:2]

        self._device.queue.write_buffer(self._param_buffer, 0, self._params)

        render_pass = command_encoder.begin_render_pass(
            color_attachments=[{
                "view": target_texture_view,
                "resolve_target": None,
                "clear_value": (0.0, 0.0, 0.0, 1.0),
                "load_op": wgpu.LoadOp.load,
                "store_op": wgpu.StoreOp.store,
            }],
        )
        render_pass.set_pipeline(self._render_pipeline)
        render_pass.set_bind_group(0, self._bind_group, [], 0, 999999)
        render_pass.set_vertex_buffer(0, self._vertex_buffer_starts)
        render_pass.set_vertex_buffer(1, self._vertex_buffer_ends)
        render_pass.draw(4, len(self._line_starts), 0, 0)
        render_pass.end()




import topsy

def test_simple_arg_parse():
    args = topsy.parse_args(["test://1000","-q","test-quantity"])
    assert len(args)==1
    args = args[0]
    assert args.filename == "test://1000"
    assert args.quantity == "test-quantity"
    assert args.resolution == topsy.config.DEFAULT_RESOLUTION
    assert args.colormap == topsy.config.DEFAULT_COLORMAP

def test_multi_arg_parse():
    args = topsy.parse_args(["file1","-q","test-quantity","-p","dm","+","file2","-q","test-quantity2"])
    assert len(args)==2

    a = args[0]
    assert a.filename == "file1"
    assert a.quantity == "test-quantity"
    assert a.particle == "dm"

    a = args[1]
    assert a.filename == "file2"
    assert a.quantity == "test-quantity2"
    assert a.particle == "dm"







from __future__ import annotations

import numpy as np
import rendercanvas.jupyter, rendercanvas.auto
import time
import copy

from .. import config

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..visualizer import Visualizer



class VisualizerCanvasBase:
    def __init__(self, *args, **kwargs):
        self._visualizer : Visualizer = kwargs.pop("visualizer")

        self._last_x = 0
        self._last_y = 0
        # The below are dummy values that will be updated by the initial resize event
        self.width_physical, self.height_physical = 640,480
        self.pixel_ratio = 1

        super().__init__(*args, **kwargs)

        self.add_event_handler(self.event_handler, "*")

    def event_handler(self, event):
        if event['event_type']=='pointer_move':
            if len(event['buttons'])>0:
                if len(event['modifiers'])==0:
                    self.drag(event['x']-self._last_x, event['y']-self._last_y)
                else:
                    self.shift_drag(event['x']-self._last_x, event['y']-self._last_y)
            self._last_x = event['x']
            self._last_y = event['y']
        elif event['event_type']=='wheel':
            self.mouse_wheel(event['dx'], event['dy'])
        elif event['event_type']=='key_up':
            self.key_up(event['key'])
        elif event['event_type']=='resize':
            self.resize_complete(event['width'], event['height'], event['pixel_ratio'])
        elif event['event_type']=='double_click':
            self.double_click(event['x'], event['y'])
        elif event['event_type']=='pointer_up':
            self.release_drag()
        else:
            pass


    def drag(self, dx, dy):
        self._visualizer.rotate(dx*0.01, dy*0.01)

    def shift_drag(self, dx, dy):
        biggest_dimension = max(self.width_physical, self.height_physical)

        displacement = 2.*self.pixel_ratio*np.array([dx, -dy, 0], dtype=np.float32) / biggest_dimension * self._visualizer.scale
        self._visualizer.position_offset += self._visualizer.rotation_matrix.T @ displacement

        self._visualizer.display_status("centre = [{:.2f}, {:.2f}, {:.2f}]".format(*self._visualizer._sph.position_offset))

        self._visualizer.crosshairs_visible = True


    def key_up(self, key):
        if key=='s':
            self._visualizer.save()
        elif key=='r':
            self._visualizer.colormap.autorange_vmin_vmax()
        elif key=='h':
            self._visualizer.reset_view()

    def mouse_wheel(self, delta_x, delta_y):
        if isinstance(self, rendercanvas.jupyter.JupyterRenderCanvas):
            # scroll events are much smaller from the web browser, for
            # some reason, compared with native windowing
            delta_y *= 10
            delta_x *= 10

        self._visualizer.scale*=np.exp(delta_y/1000)

    def release_drag(self):
        if self._visualizer.crosshairs_visible:
            self._visualizer.crosshairs_visible = False
            self._visualizer.invalidate()


    def resize(self, *args):
        # putting this here as a reminder that the resize method must be passed to the base class
        super().resize(*args)

    def resize_complete(self, width, height, pixel_ratio=1):
        self.width_physical = int(width*pixel_ratio)
        self.height_physical = int(height*pixel_ratio)
        self.pixel_ratio = pixel_ratio

    def double_click(self, x, y):
        original_position = copy.copy(self._visualizer.position_offset)

        biggest_dimension = max(self.width_physical, self.height_physical)


        centre_physical_x = self.width_physical / (2*self.pixel_ratio)
        centre_physical_y = self.height_physical / (2*self.pixel_ratio)

        xy_displacement = 2. * self.pixel_ratio * np.array([centre_physical_x-x,
                                                            y-centre_physical_y,
                                                            0], dtype=np.float32) / biggest_dimension * self._visualizer.scale


        self._visualizer.position_offset += self._visualizer.rotation_matrix.T @ xy_displacement


        depth_im = self._visualizer.get_depth_image()
        central_depth = depth_im[depth_im.shape[0]//2, depth_im.shape[1]//2]

        if ~np.isnan(central_depth):
            z_displacement = np.array([0, 0, -central_depth], dtype=np.float32)
            self._visualizer.position_offset += self._visualizer.rotation_matrix.T @ z_displacement

        final_position = self._visualizer.position_offset

        # the actual work is done - now animate it so it looks understandable
        self._visualizer.position_offset = original_position

        #def interpolate_position(t):
        #    return original_position + (final_position - original_position) * t

        def interpolate_position(t):
            w1 = np.arctan(5*(t*2-1))/np.pi+0.5
            w2 = 1-w1
            return w2 * original_position + w1 * final_position

        start = time.time()

        def glide():
            t = (time.time()-start)/config.GLIDE_TIME
            if t>1:
                self._visualizer.position_offset = final_position
            else:
                self.call_later(0.0, glide)
                self._visualizer.position_offset = interpolate_position(t)



        self.call_later(1. / config.TARGET_FPS, glide)


    @classmethod
    def call_later(cls, delay, fn, *args):
        raise NotImplementedError()





# Now we are going to select a specific backend
#
# we don't use rendercanvas.auto directly because it prefers the glfw backend over qt
# whereas we want to use qt
#
# Note also that is_jupyter as implemented fails to distinguish correctly if we are
# running inside a kernel that isn't attached to a notebook. There doesn't seem to
# be any way to distinguish this, so we live with it for now.


from .. import is_jupyter

if is_jupyter():
    from .jupyter import VisualizerCanvas
else:
    from .qt import VisualizerCanvas




"""Interpolation between frames for the motion recorder"""

import math
import numpy as np

from abc import ABC, abstractmethod

class Interpolator(ABC):
    """ABC for interpolating a timestream.

    The timestream is a list of (time, value) pairs, where time is a float and value is any type."""

    no_value = object()

    def __init__(self, timestream):
        self._timestream = timestream

    @abstractmethod
    def __call__(self, t):
        pass


class LinearInterpolator(Interpolator):
    """Returns the linearly interpolated value, or None if no value is available"""

    def __call__(self, t):
        stream = self._timestream
        for i, (t_ev, val_ev) in enumerate(stream):
            if t_ev >= t:
                if i == 0:
                    return val_ev
                else:
                    t0, val0 = stream[i - 1]
                    assert t0 < t
                    return val0 + (val_ev - val0) * (t - t0) / (t_ev - t0)
        return self.no_value

class SmoothedInterpolatorMixin:
    def __init__(self, timestream, smoothing=0.25, fps=30):
        """Create a linear interpolator with gaussian smoothing over the specified period

        Args:
            timestream: the timestream to interpolate
            smoothing: the standard deviation of the gaussian smoothing kernel, in seconds
            fps: the number of samples per second in the smoothed timestream (doesn't have to match video fps)
        """
        super().__init__(timestream)
        tmax = timestream[-1][0]
        self._smoothing = smoothing

        interpolated_timestream = []
        for i in range(math.floor(tmax*fps)):
            interpolated_timestream.append(super().__call__(i/fps))


        kernel = np.exp(-np.arange(-3*smoothing*fps, 3*smoothing*fps)**2/(2*smoothing**2*fps**2))
        kernel/=kernel.sum()
        interpolated_timestream = np.concatenate(
            ([interpolated_timestream[0]]*(len(kernel)//2),
             interpolated_timestream,
             [interpolated_timestream[-1]]*(len(kernel)//2))
        )

        if len(interpolated_timestream.shape)==1:
            smoothed_timestream = np.convolve(interpolated_timestream, kernel, mode='valid')
        else:
            smoothed_timestream = None
            for index in np.ndindex(interpolated_timestream.shape[1:]):
                index_c = (slice(None),)+index # py3.11+ supports [:, *index] but not py3.10-
                result = np.convolve(interpolated_timestream[index_c], kernel, mode='valid')
                if smoothed_timestream is None:
                    smoothed_timestream = np.empty((len(result),)+interpolated_timestream.shape[1:])
                smoothed_timestream[index_c] = result

        self._timestream = [(i/fps, val) for i, val in enumerate(smoothed_timestream)]


class SmoothedLinearInterpolator(SmoothedInterpolatorMixin, LinearInterpolator):
    pass

class RotationInterpolator(LinearInterpolator):
    """Returns an interpolated rotation matrix"""

    def __call__(self, t):
        matr = super().__call__(t)
        if matr is self.no_value:
            return matr

        # orthogonalise matr:
        u, s, vh = np.linalg.svd(matr)
        return u @ vh

class SmoothedRotationInterpolator(SmoothedInterpolatorMixin, RotationInterpolator):
    pass


class StepInterpolator(Interpolator):
    """Only returns a value when it has changed. Assumes it is being acccessed sequentially"""

    def __init__(self, timestream):
        super().__init__(timestream)
        self._last_value = self.no_value
        self._last_t = None

    def __call__(self, t):
        if self._last_t is not None and t<self._last_t:
            raise ValueError("StepInterpolator must be accessed sequentially")

        self._last_t = t

        stream = self._timestream
        for (t_ev, val_ev) in stream[::-1]:
            if t_ev <= t:
                if val_ev != self._last_value:
                    self._last_value = val_ev
                    return self._last_value
                else:
                    return self.no_value
        return self.no_value

class SmoothedStepInterpolator(StepInterpolator):
    def __init__(self, timestream, smoothing=0.25):
        self._start_value = None
        self._target_value = None
        self._transition_start = None
        self._transition_end = None
        self._smoothing = smoothing
        super().__init__(timestream)

    def __call__(self, t):
        if self._target_value is not None:
            if t>=self._transition_end:
                tv = self._target_value
                self._start_value = None
                self._target_value = None
                self._transition_start = None
                self._transition_end = None
                return tv
            else:
                return self._start_value + (self._target_value-self._start_value)*(t-self._transition_start)/(self._transition_end-self._transition_start)
        else:
            last_value = self._last_value
            new_value = super().__call__(t)
            if new_value is self.no_value or new_value == last_value:
                return self.no_value
            elif last_value is self.no_value:
                return new_value
            else:
                self._start_value = last_value
                self._target_value = new_value
                self._transition_start = t
                self._transition_end = t + self._smoothing
                return last_value




"""topsy - An astrophysics simulation visualization package based on webgpu, using pynbody for reading data"""

from __future__ import annotations

__version__ = "0.6.2"

import argparse
import logging
import sys

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import pynbody


from . import config

logger = None

def parse_args(args=None):
    """Create arguments and kwargs to pass to the visualizer, from sys.argv"""

    # create the argument parser, add arguments for filename, resolution, and colormap, and parse the arguments
    argparser = argparse.ArgumentParser(description="Visualize an astrophysics simulation. Multiple windows can be opened by separating groups of arguments with *.")

    argparser.add_argument("filename", help="Specify path to a simulation file to be visualized")
    argparser.add_argument("--resolution", "-r", help="Specify the resolution of the visualization",
                           default=config.DEFAULT_RESOLUTION, type=int)
    argparser.add_argument("--colormap", "-m", help="Specify the matplotlib colormap to be used",
                           default=config.DEFAULT_COLORMAP, type=str)
    argparser.add_argument("--particle", "-p", help="Specify the particle type to visualise",
                            default="dm", type=str)
    argparser.add_argument("--center", "-c", help="Specify the centering method: 'halo-<N>', 'all', 'zoom' or 'none'",
                           default="none", type=str)
    argparser.add_argument("--quantity", "-q", help="Specify a quantity to render instead of density",
                           default=None, type=str)
    argparser.add_argument("--tile", "-t", help="Wrap and tile the simulation box using its periodicity",
                           default=False, action="store_true")
    argparser.add_argument('--hdr', help="[Experimental] Enable HDR rendering", action="store_true")
    argparser.add_argument('--rgb', help="[Experimental] Enable RGB->UVI rendering for stars", action="store_true")

    argparser.add_argument("--load-sphere", nargs='+', help="Load a sphere of particles with the given "
                                                          "radius and, optionally, centre in simulation units. "
                                                          "e.g. --load-sphere 5.0 to load a sphere of radius 5.0 about"
                                                          "the centre of the simulation, or 5.0 3.0 1.0 2.0 to load a "
                                                          "sphere of radius 5.0 about the point (3.0, 1.0, 2.0)."
                                                          "Supported only for swift simulations. Units are simulation units.",
                            metavar=("_"),
                            default=None, type=float)

    if args is None:
        args = sys.argv[1:]
    arg_batches = []
    # split args into batches separated by '+'
    while len(args) > 0:
        try:
            split_index = args.index("+")
        except ValueError:
            split_index = len(args)

        this_args = argparser.parse_args(args[:split_index])

        if this_args.load_sphere is not None and len(this_args.load_sphere) != 1 and len(this_args.load_sphere) != 4:
                argparser.error("Invalid number of arguments for --load-sphere. Must be 1 or 4.")
        arg_batches.append(this_args)
        args = args[split_index+1:]


    return arg_batches

def setup_logging():
    global logger
    if logger is not None:
        return
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )
    logger.addHandler(ch)

def main():
    all_args = parse_args()

    for args in all_args:
        vis = load(args.filename, center=args.center, resolution=args.resolution,
                    particle=args.particle, tile=args.tile, rgb=args.rgb,
                    sphere_radius=args.load_sphere[0] if args.load_sphere is not None else None,
                    sphere_center=tuple(args.load_sphere[1:]) if args.load_sphere is not None and len(args.load_sphere) == 4 else None,
                    hdr=args.hdr)
        vis.quantity_name = args.quantity
        vis.canvas.show()

    from rendercanvas import qt # has to be imported here so that underlying qt toolkit has been autoselected
    qt.loop.run()

def topsy(snapshot: pynbody.snapshot.SimSnap, quantity: str | None = None, **kwargs):
    from . import visualizer, loader
    vis = visualizer.Visualizer(data_loader_class=loader.PynbodyDataInMemory,
                                data_loader_args=(snapshot,),
                                **kwargs)
    vis.quantity_name = quantity
    return vis

def load(filename: str, center: str = "none", particle: str = "gas", rgb: bool = False,
         resolution: int = config.DEFAULT_RESOLUTION, tile: bool = False,
         sphere_radius: float | None = None, sphere_center: tuple[float, float, float] | None = None,
         hdr: bool = False) :
    """
    Load a simulation file (currently using pynbody) and return a visualizer object.

    Parameters
    ----------

    filename : str
        Path to the simulation file. You can also specify test://<N> to generate a test dataset with N particles.

    center : str
        Centering method. Can be 'halo-<N>', 'all', 'zoom' or 'none'.

    particle : str
        Particle type to visualize. Default is 'gas'; other options include 'dm' and 'star'.

    resolution : int
        Resolution of the visualization in pixels.

    sphere_radius : float | None
        If specified, load a sphere of particles with the given radius. Units are simulation units.

    sphere_center : tuple[float, float, float] | None
        If specified, load a sphere of particles with the given center. Units are simulation units.
        Must be a tuple of three floats (x, y, z).

    rgb : bool
        If True, enable RGB->UVI rendering for stars. Default is False.

    hdr : bool
        If True, try enabling HDR rendering (only valid when rgb=True). Default is False.

    tile : bool
        If True, wrap and tile the simulation box using its periodicity. Default is False.

    Returns
    -------
    visualizer.Visualizer
        A visualizer object that can be used to render the simulation data.

    """
    from . import visualizer, loader
    setup_logging()
    
    if "test://" in filename:
        loader_class = loader.TestDataLoader
        try:
            n_part = int(float(filename[7:]))  # going through float allows scientific notation
        except ValueError:
            n_part = config.TEST_DATA_NUM_PARTICLES_DEFAULT
        logger.info(f"Using test data with {n_part} particles")
        loader_args = (n_part,)
    else:
        import pynbody
        loader_class = loader.PynbodyDataLoader
        if sphere_radius is not None:
            if sphere_center is not None:
                loader_args = (filename, center, particle, pynbody.filt.Sphere(sphere_radius, sphere_center))
            else:
                loader_args = (filename, center, particle, pynbody.filt.Sphere(sphere_radius))
        else:
            loader_args = (filename, center, particle)

    vis = visualizer.Visualizer(data_loader_class=loader_class,
                                data_loader_args=loader_args,
                                hdr=hdr,
                                periodic_tiling=tile,
                                render_resolution=resolution,
                                rgb=rgb)

    return vis

def test(nparticle=config.TEST_DATA_NUM_PARTICLES_DEFAULT, **kwargs):
    from . import visualizer, loader
    vis = visualizer.Visualizer(data_loader_class=loader.TestDataLoader,
                                data_loader_args=(nparticle,),
                                data_loader_kwargs={'with_cells': kwargs.pop('with_cells', False),
                                                    'periodic': kwargs.get('periodic_tiling', False)},
                                **kwargs)
    return vis




_force_is_jupyter = False

def is_jupyter():
    """Determine whether the user is executing in a Jupyter Notebook / Lab.

    This has been pasted from an old version of wgpu.gui.auto.is_jupyter; the function was removed"""
    global _force_is_jupyter
    if _force_is_jupyter:
        return True
    from IPython import get_ipython
    try:
        ip = get_ipython()
        if ip is None:
            return False
        if ip.has_trait("kernel"):
            return True
        else:
            return False
    except NameError:
        return False
    
def force_jupyter():
    """Force the return from is_jupyter() to be True; used in testing"""
    global _force_is_jupyter
    _force_is_jupyter = True



DEFAULT_RESOLUTION = 1024
DEFAULT_COLORMAP = 'twilight_shifted'

DEFAULT_SCALE = 200.0 # viewport width in kpc

TARGET_FPS = 30 # will use downsampling to achieve this
INITIAL_PARTICLES_TO_RENDER = 1e5 # number of particles to render at first
STATUS_LINE_UPDATE_INTERVAL = 0.2 # seconds
STATUS_LINE_UPDATE_INTERVAL_RAPID = 0.05 # when time-critical information is being displayed

GLIDE_TIME = 0.3 # seconds after double click to reach destination

COLORBAR_ASPECT_RATIO = 0.15
COLORMAP_NUM_SAMPLES = 1000

TEST_DATA_NUM_PARTICLES_DEFAULT = int(1e6)

MAX_PARTICLES_PER_BUFFER = 2**27
# arbitrary number, but small enough that GPU memory fragmentation not a huge issue hopefully, while
# large enough to not cause too much overhead

MAX_PARTICLES_PER_EXPORT_RENDERCALL = 2 ** 25
# again, a arbitraryish number, determined by experimentation. Profiling finds that calling render on a large number of particles
# causes a lot of overhead - perhaps due to pipeline stalls within the GPU. This becomes particularly
# noticeable on EXPORT rendering large simulations. Splitting into a sequence of smaller calls.

DEFAULT_CELLS_NSIDE = 16
# To provide a quick way to remove unneeded verticies, the simulation is ordered into cells
# by default, use this number of cells on a side (so DEFAULT_CELLS_NSIDE^3 in total).
# High numbers enable more precise geometric selections, but at a general performance penalty
# due to the complexity of the GPU buffer organization.

JUPYTER_UI_LAG = 0.05
# time over which to spread jupyter UI updates, notably for sliders where updating the range and value
# simultaneously seems to lead to problems

# special name for  projected density in UI
PROJECTED_DENSITY_NAME = "Projected density"


import math
import numpy as np

from . import config
from . import drawreason
from .cell_layout import CellLayout

class RenderProgression:
    """Recommends the block of particles to SPH render, taking into account historical timing"""
    def __init__(self, total_particles, initial_particles = None):
        if initial_particles is None:
            initial_particles = int(config.INITIAL_PARTICLES_TO_RENDER)
        self._recommended_num_particles_to_render = min(initial_particles, total_particles)
        self._start_index = 0
        self._max_num_particles = total_particles
        self._current_draw_reason = None
        self._last_num_to_render = 1

    def get_max_particle_regions_per_block(self):
        """Get the maximum number of particle regions that will be returned by get_block"""
        return 1

    def start_frame(self, draw_reason: drawreason.DrawReason):
        """Called at the start of a frame to reset the start index if needed

        Returns
        -------

        update_particle_ranges: bool
            Whether the particle ranges need to be updated before render command is issued

        """
        self._current_draw_reason = draw_reason
        self._first_block_in_frame = True
        self._total_num_rendered_in_frame = 0
        if draw_reason not in (drawreason.DrawReason.PRESENTATION_CHANGE, drawreason.DrawReason.REFINE):
            self._start_index = 0
            return True
        else:
            return False

    def end_frame_get_scalefactor(self):
        """Ends a frame and returns the scale factor for the colormap"""
        self._perform_particle_number_update()
        self._current_draw_reason = None
        return self._max_num_particles / self._start_index

    def get_block(self, time_elapsed_in_frame: float) -> tuple[list[int], list[int]] | None:
        """Returns a list of starting indicies and lengths from the particle buffer to render"""
        if self._current_draw_reason is None:
            raise RuntimeError("get_block called without a current frame")
        draw_reason = self._current_draw_reason
        if draw_reason == drawreason.DrawReason.PRESENTATION_CHANGE:
            return None
        elif draw_reason == drawreason.DrawReason.EXPORT:
            if self._start_index < self._max_num_particles:
                try_rendering = self._max_num_particles - self._start_index
                max_to_render = int(config.MAX_PARTICLES_PER_EXPORT_RENDERCALL / self.get_fraction_volume_selected())
                if try_rendering > max_to_render:
                    try_rendering = max_to_render
                self._last_num_to_render = try_rendering
                return ([self._start_index], [try_rendering])
            else:
                return None
        else:
            if self._start_index >= self._max_num_particles:
                return None

            if self._first_block_in_frame:
                time_available = 1./config.TARGET_FPS
                self._first_block_in_frame = False
            else:
                time_available = 1./config.TARGET_FPS - time_elapsed_in_frame

            if time_available<=0.4/config.TARGET_FPS:
                return None
                # while we could set a time_available based on the time taken to render the last block here, we
                # no longer attempt to do multiple blocks per frame in interactive context,
                # as it leads to unpredictable frame rates and stuttering. Instead, we will just render one block per
                # frame. We'll correct any mismatch in particle number on the next frame (possibly a REFINE frame).

            num_to_render = int(self._recommended_num_particles_to_render * time_available * config.TARGET_FPS)
            if num_to_render + self._start_index > self._max_num_particles:
                num_to_render = self._max_num_particles - self._start_index
            self._last_num_to_render = num_to_render
            return ([self._start_index], [num_to_render])

    def _perform_particle_number_update(self):
        num_achievable = int(self._total_num_rendered_in_frame / (self._time_in_frame * config.TARGET_FPS))
        num_achievable = min(num_achievable, self._max_num_particles)
        if num_achievable < 1:
            # very strange edge case, but must never recommend rendering less than one particle!
            num_achievable = 1

        if self._current_draw_reason != drawreason.DrawReason.REFINE:
            log2_change = abs(math.log2(num_achievable) - math.log2(self._recommended_num_particles_to_render))
            if log2_change > 1.5:
                # emergency situation, update completely
                self._recommended_num_particles_to_render = num_achievable
            elif log2_change > 0.3:
                # modest mismatch, make a more cautious update
                self._recommended_num_particles_to_render = int(
                    num_achievable ** 0.3 * self._recommended_num_particles_to_render ** 0.7)


    def end_block(self, time_elapsed_in_frame: float):
        """Report the time taken to render a number of particles, so that the next recommendation can be made"""
        self._start_index += self._last_num_to_render
        self._total_num_rendered_in_frame += self._last_num_to_render
        self._time_in_frame = time_elapsed_in_frame

        """ if self._current_draw_reason == drawreason.DrawReason.REFINE:
            # the next frame needs to update the GPU render particle ranges, come what may
            self._update_particle_ranges = True
        elif (abs(math.log2(num_achievable) - math.log2(self._recommended_num_particles_to_render)) > 0.6):
            if was_first_block_in_frame:
                # only update based on the first block in the frame (which is the one which is most critical to
                # get to roughly the right length)
                self._recommended_num_particles_to_render = num_achievable
                self._update_particle_ranges = True
        else:
            self._update_particle_ranges = False"""

    def needs_refine(self):
        """Check if the render progression is not yet complete"""
        needs_refine = self._start_index < self._max_num_particles
        return needs_refine

    def select_sphere(self, cen, radius):
        # default render progression has no way to select a sphere of particles
        pass

    def select_all(self):
        pass

    def get_fraction_volume_selected(self):
        return 1.0


class RenderProgressionWithCells(RenderProgression):
    def __init__(self, cell_layout: CellLayout, total_particles: int, initial_particles=None):
        super().__init__(total_particles, initial_particles)
        self._cell_layout = cell_layout
        random_state = np.random.RandomState(1337)
        self._cell_phase_shifts = random_state.permutation(self._cell_layout.get_num_cells())
        self._selected_cells_hash = 0
        self.select_all()

    def get_max_particle_regions_per_block(self):
        return self._cell_layout.get_num_cells()

    def _map_logical_range_to_actual_ranges(self, start, length):
        """Map from logical range to actual ranges in the cell layout.

        Performance critical - called during rendering. Optimized into numpy array operations but could
        probably be usefully optimized further based on profiling. (Although should double check
        in the profile that it really is *this* routine that is taking the time.)
        """
        num_particles = self._cell_layout.get_num_particles()
        fractional_start = start / num_particles
        fractional_end = (start + length) / num_particles

        num_cells = self._cell_layout.get_num_cells()
        offset_per_cell = self._cell_layout._offsets

        # the 'phase shift' ensures that if a very low number of particles are selected such that the mean
        # number of particles per cell is less than one, some particles still get selected when we are
        # starting at zero. Otherwise quantization effects would make it impossible to select any particles
        # until a much later block. Also the phase shift must be evenly distributed across cell so that
        # we don't get differential spatial effects
        cell_phase_shifts = self._cell_phase_shifts/num_cells
        total_particles_in_cells = self._cell_layout._lengths

        ideal_start_per_cell = fractional_start*total_particles_in_cells.astype(np.float64)
        ideal_end_per_cell = fractional_end*total_particles_in_cells.astype(np.float64)

        # the above are floating point numbers, but we can actually only take an integer, so floor it
        start_per_cell = (ideal_start_per_cell+cell_phase_shifts).astype(np.intp)
        end_per_cell = (ideal_end_per_cell+cell_phase_shifts).astype(np.intp)
        len_per_cell = end_per_cell-start_per_cell

        start_global = (start_per_cell + offset_per_cell)[self._selected_cells]
        len_global = len_per_cell[self._selected_cells]

        mask = len_global>0

        return start_global[mask], len_global[mask]



    def get_block(self, time_elapsed_in_frame: float) -> tuple[list[int], list[int]] | None:
        result = super().get_block(time_elapsed_in_frame)
        if result is None:
            return None
        starts, lens = result
        assert len(starts) == len(lens) == 1
        if lens[0] == self._max_num_particles:
            return starts, lens
        else:
            return self._map_logical_range_to_actual_ranges(starts[0], lens[0])

    def select_all(self):
        """Select all cells for inclusion in next render pass"""
        self._selected_cells = np.arange(self._cell_layout.get_num_cells())
        self._check_cells_for_update()

    def select_sphere(self, cen, r):
        """Select a sphere of particles for inclusion in next render pass"""
        self._selected_cells = self._cell_layout.cells_in_sphere(cen, r)
        self._check_cells_for_update()

    def _check_cells_for_update(self):
        sc_hash = hash(self._selected_cells.tobytes())
        if sc_hash != self._selected_cells_hash:
            self._selected_cells_hash = sc_hash
            self._update_particle_ranges = True

    def get_fraction_volume_selected(self):
        """Get the number of cells selected for inclusion in next render pass"""
        return len(self._selected_cells)/self._cell_layout.get_num_cells()







import logging
import numpy as np
import wgpu

from . import config, performance

logger = logging.getLogger(__name__)

class SplitBuffers:
    """Manages splitting buffers into smaller buffers for GPU hardware.

    This is needed because some/most GPUs have a limit on the size of buffers which is less than the actual total
    RAM available.

    We adopt terminology of a 'global' address (in terms of particles) which can then be mapped onto a 'split'
    address which is a tuple of (buffer number, buffer offset)."""

    def __init__(self, num_particles: int, max_particles_per_buffer: int | None = None):
        if max_particles_per_buffer is None:
            max_particles_per_buffer = config.MAX_PARTICLES_PER_BUFFER

        self._num_particles = num_particles
        self._max_particles_per_buffer = max_particles_per_buffer
        self._calculate_splits()

    def _calculate_splits(self):
        if self._num_particles > self._max_particles_per_buffer:
            num_buffers = int(np.ceil(self._num_particles / self._max_particles_per_buffer))
        else:
            num_buffers = 1

        self._num_buffers = num_buffers
        self._buffer_particle_sizes = np.empty(num_buffers, dtype=np.intp)
        self._buffer_particle_sizes.fill(self._max_particles_per_buffer)
        self._buffer_particle_sizes[-1] = self._num_particles - (len(self._buffer_particle_sizes)-1) * self._max_particles_per_buffer

        self._buffer_particle_starts = np.cumsum(self._buffer_particle_sizes) - self._buffer_particle_sizes
        logger.info(f"Splitting {self._num_particles} particles into {self._num_buffers} buffer(s)")


    def _global_to_split_address(self, address: int) -> (int, int):
        """Given a logical buffer particle offset, returns the physical buffer number and address"""
        bufnum = np.searchsorted(self._buffer_particle_starts, address, side='right')-1
        return bufnum, address - self._buffer_particle_starts[bufnum]

    @property
    def num_buffers(self) -> int:
        """Number of buffers per global buffer"""
        return self._num_buffers

    def global_to_split(self, start: int, length: int) -> tuple[list[int], list[int], list[int]]:
        """Map global start and length to split buffer numbers, starts and lengths."""
        bufs = []
        starts = []
        lengths = []

        global_start = start
        global_length_remaining = length
        bufnum, local_start = self._global_to_split_address(global_start)

        while global_length_remaining>0 and bufnum<self._num_buffers:
            maxlen = self._buffer_particle_sizes[bufnum] - local_start
            buf_len = min(global_length_remaining, maxlen)
            bufs.append(bufnum)
            starts.append(local_start)
            lengths.append(buf_len)

            global_start += buf_len
            global_length_remaining -= buf_len
            bufnum += 1
            local_start = 0

        if global_length_remaining > 0:
            raise ValueError(f"Requested length {length} starting at {start} exceeds available buffers")

        return bufs, starts, lengths

    def global_to_split_monotonic(self, start: list[int], length: list[int]) -> list[tuple[list[int], list[int]]]:
        """Map global start and length to starts and lengths for each buffer. Addressing must be monotonically increasing."""
        performance.signposter.emit_event("global_to_split_monotonic")
        starts = []
        lengths = []

        cur_buf = 0
        cur_buf_start = 0
        cur_buf_end = self._buffer_particle_sizes[cur_buf]

        all_buf_results = [(starts, lengths)]

        for global_start, global_length in zip(start, length):
            while global_length > 0:
                while global_start >= cur_buf_end:
                    # move to next buffer
                    cur_buf += 1
                    if cur_buf>=self._num_buffers:
                        raise ValueError(f"Requested length {global_length} starting at {global_start} exceeds available buffers")
                    cur_buf_start = self._buffer_particle_starts[cur_buf]
                    cur_buf_end = cur_buf_start + self._buffer_particle_sizes[cur_buf]
                    starts = []
                    lengths = []
                    all_buf_results.append((starts, lengths))

                this_buf_start = global_start - cur_buf_start
                this_buf_length = min(global_length, cur_buf_end - global_start)
                starts.append(this_buf_start)
                lengths.append(this_buf_length)
                global_length -= this_buf_length
                global_start += this_buf_length

        if cur_buf < self._num_buffers-1:
            for bufnum in range(cur_buf+1, self._num_buffers):
                all_buf_results.append(([], []))

        performance.signposter.emit_event("end global_to_split_monotonic")

        return all_buf_results


    def create_buffers(self, wgpu_device: wgpu.GPUDevice, item_size: int, usage: wgpu.BufferUsage) -> list[wgpu.GPUBuffer]:
        """Create a set of split buffers

        Parameters
        ----------
        wgpu_device : wgpu.GPUDevice
            The GPU device to create the buffer on.
        item_size : int
            The size of each item in the buffer.
        usage : int
            The usage flags for the buffer.

        """
        buffers = []
        for this_size in self._buffer_particle_sizes:
            size = this_size * item_size
            buffer = wgpu_device.create_buffer(
                size=size,
                usage=usage,
            )
            buffers.append(buffer)
        return buffers

    def write_buffers(self, wgpu_device: wgpu.GPUDevice, buffers: list[wgpu.GPUBuffer], data: np.ndarray) -> None:
        """Write data to the split buffers.

        Parameters
        ----------
        wgpu_device : wgpu.GPUDevice
            The GPU device to write the buffer on.
        buffers : list[wgpu.GPUBuffer]
            The buffers to write to.
        data : np.ndarray
            The data to write.
        """
        if len(buffers) != self._num_buffers:
            raise ValueError(f"Number of buffers {len(buffers)} does not match number of split buffers {self._num_buffers}")
        if len(data) != self._num_particles:
            raise ValueError(f"Data size {len(data)} does not match number of particles {self._num_particles}")

        for bufnum, buf in enumerate(buffers):
            start = self._buffer_particle_starts[bufnum]
            length = self._buffer_particle_sizes[bufnum]
            wgpu_device.queue.write_buffer(buf, 0, data[start:start+length])


import numpy as np
import pytest
import wgpu

from topsy import split_buffers

@pytest.fixture
def device():
    adapter = wgpu.gpu.request_adapter_sync(power_preference="high-performance")
    return adapter.request_device_sync()

@pytest.fixture
def sb(max_per_buffer=15, num_particles=50):
    return split_buffers.SplitBuffers(num_particles, max_per_buffer)

def test_global_to_split(sb):

    assert sb.global_to_split(0, 10) == ([0], [0], [10])
    assert sb.global_to_split(0, 20) == ([0, 1], [0, 0], [15, 5])
    assert sb.global_to_split(0, 45) == ([0, 1, 2], [0, 0, 0], [15, 15, 15])
    assert sb.global_to_split(15, 10) == ([1], [0], [10])
    assert sb.global_to_split(14, 2) == ([0, 1], [14, 0], [1, 1])
    assert sb.global_to_split(20, 20) == ([1, 2], [5, 0], [10, 10])
    assert sb.global_to_split(49, 1) == ([3], [4], [1])
    assert sb.global_to_split(0, 50) == ([0, 1, 2, 3], [0, 0, 0, 0], [15, 15, 15, 5])

    with pytest.raises(ValueError):
        # too long:
        sb.global_to_split(0, 100)

    with pytest.raises(ValueError):
        sb.global_to_split(49,2)

def test_global_to_split_monotonic(sb):
    def generate_test_case():
        starts = np.random.randint(0, 50, size=5)
        starts = np.sort(starts)
        lengths = np.diff(starts)
        starts = starts[:-1]
        lengths = np.random.randint(lengths+1)

        # remove any cases i with lengths[i] = 0 from the starts and lengths lists:
        mask = lengths!=0
        starts = starts[mask]
        lengths = lengths[mask]

        return (starts, lengths)

    np.random.seed(1337)
    for i in range(100):
        starts, lengths = generate_test_case()

        # faster version to be used in code:
        results = sb.global_to_split_monotonic(starts, lengths)

        # slow version to act as check:
        results_slow = [([], []) for _ in range(sb._num_buffers)]

        for global_s, global_l in zip(starts, lengths):
            local_bufs, local_starts, local_lengths = sb.global_to_split(global_s, global_l)
            for lb, ls, ll in zip(local_bufs, local_starts, local_lengths):
                results_slow[lb][0].append(ls)
                results_slow[lb][1].append(ll)

        assert results == results_slow


def test_create_buffers(device, sb):
    # Create a buffer with 4 buffers of size 15 and one of size 5
    buffers = sb.create_buffers(device, 4, wgpu.BufferUsage.UNIFORM)
    assert len(buffers) == 4
    assert all([buf.size == 15*4 for buf in buffers[:-1]])
    assert buffers[-1].size == 5*4

def test_write_buffers(device, sb):
    buffers = sb.create_buffers(device, 4, wgpu.BufferUsage.UNIFORM | wgpu.BufferUsage.COPY_DST)

    data = np.arange(50, dtype=np.float32)

    with pytest.raises(ValueError):
        # wrong number of buffers
        sb.write_buffers(device, buffers[:-1], data)

    with pytest.raises(ValueError):
        # wrong number of particles
        sb.write_buffers(device, buffers, data[:-1])

    # should succeed
    sb.write_buffers(device, buffers, data)


from __future__ import annotations

from rendercanvas.offscreen import OffscreenRenderCanvas, loop

from . import VisualizerCanvasBase


class VisualizerCanvas(VisualizerCanvasBase, OffscreenRenderCanvas):

    @classmethod
    def call_later(cls, delay, fn, *args):
        loop.call_later(delay, fn, *args)




import numpy as np
import wgpu
import logging
import pynbody
import pickle

from typing import Optional

from . import config, cell_layout

from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)


class AbstractDataLoader(ABC):
    def __init__(self, device: wgpu.GPUDevice):
        self._device = device

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def get_positions(self):
        pass

    @abstractmethod
    def get_smooth(self):
        pass

    @abstractmethod
    def get_mass(self):
        pass

    @abstractmethod
    def get_named_quantity(self, name):
        pass

    @abstractmethod
    def get_quantity_label(self, quantity_name):
        pass

    @abstractmethod
    def get_rgb_masses(self):
        pass

    def get_pos_smooth(self):
        pos_smooth = np.empty((len(self), 4), dtype=np.float32)
        pos_smooth[:, :3] = self.get_positions()
        pos_smooth[:, 3] = self.get_smooth()
        return pos_smooth

    def get_periodicity_scale(self):
        return np.inf

    def get_render_progression(self):
        from . import progressive_render
        if hasattr(self, '_cell_layout'):
            return progressive_render.RenderProgressionWithCells(self._cell_layout, len(self))
        else:
            return progressive_render.RenderProgression(len(self))


class PynbodyDataInMemory(AbstractDataLoader):
    """Base class for data loaders that use pynbody."""

    _name_smooth_array = 'smooth'

    def __init__(self, device: wgpu.GPUDevice, snapshot: pynbody.snapshot.SimSnap):
        super().__init__(device)

        self.snapshot = snapshot

        boxmin = self.snapshot['pos'].min()-1e-6
        boxmax = self.snapshot['pos'].max()+1e-6
        self._cell_layout, ordering = cell_layout.CellLayout.from_positions(self.snapshot['pos'], boxmin, boxmax,
                                                                            config.DEFAULT_CELLS_NSIDE)
        self._particle_order = ordering[self._cell_layout.randomize_within_cells()]

    def get_positions(self):
        return self.snapshot['pos'].astype(np.float32)[self._particle_order]

    def get_smooth(self):
        return self.snapshot[self._name_smooth_array].astype(np.float32)[self._particle_order]

    def get_mass(self):
        return self.snapshot['mass'].astype(np.float32)[self._particle_order]

    def _effective_mass_for_band(self, band):
        return (10 ** (-0.4 * self.snapshot[band + "_mag"]))[self._particle_order]

    def get_rgb_masses(self):
        rgb = np.empty((len(self.snapshot), 3), dtype=np.float32)
        rgb[:, 0] = self._effective_mass_for_band('I') * 0.5
        rgb[:, 1] = self._effective_mass_for_band('V')
        rgb[:, 2] = self._effective_mass_for_band('U')
        rgb[np.isnan(rgb)] = 0.0
        return rgb

    def get_named_quantity(self, name):
        qty = self.snapshot[name]
        if len(qty.shape) == 2:
            qty = qty[:, 0]
        return qty.astype(np.float32)[self._particle_order]

    def get_quantity_names(self):
        return self.snapshot.loadable_keys()

    def get_quantity_label(self, quantity_name):
        if quantity_name is None:
            return r"density / $M_{\odot} / \mathrm{kpc}^2$"
        else:
            lunit = self.snapshot[quantity_name].units.latex()
            if lunit != "":
                lunit = "$/" + lunit + "$"
            return quantity_name + lunit

    def __len__(self):
        return len(self.snapshot)

    def get_periodicity_scale(self):
        if 'boxsize' in self.snapshot.properties:
            return float(self.snapshot.properties['boxsize'].in_units("kpc"))
        else:
            return None

    def get_filename(self):
        return self.snapshot.filename


class PynbodyDataLoader(PynbodyDataInMemory):
    """Literal data loader for pynbody (starts from just a filename)"""

    _name_smooth_array = 'topsy_smooth'

    def __init__(self, device: wgpu.GPUDevice, filename: str, center: str, particle: str,
                 take_region: Optional[pynbody.filt.Filter] = None):

        logger.info(f"Data filename = {filename}, center = {center}, particle = {particle}")
        if take_region is None:
            snapshot = pynbody.load(filename)
        else:
            snapshot = pynbody.load(filename, take_region=take_region)

        snapshot.physical_units()
        self.filename = filename

        fam = pynbody.family.get_family(particle)
        snapshot = snapshot[fam]

        self._family_name = fam.name
        logger.info("Loading position data...")
        _ = snapshot['pos'] # just trigger the load
        self.snapshot = snapshot

        self._perform_centering(center)

        super().__init__(device, snapshot)

        self._perform_smoothing()

    @property
    def _smooth_cache_filename(self):
        return f"{self.filename}-topsy-smooth-{self._family_name}.pkl"

    @property
    def _rho_cache_filename(self):
        return f"{self.filename}-topsy-rho-{self._family_name}.pkl"

    def _perform_centering(self, center):
        logger.info("Performing centering...")
        if center.startswith("halo-"):
            halo_number = int(center[5:])
            h = self.snapshot.ancestor.halos()
            pynbody.analysis.halo.center(h[halo_number], vel=False)
            self.snapshot.wrap()
        elif center == 'zoom':
            f_dm = self.snapshot.ancestor.dm
            pynbody.analysis.halo.center(f_dm[f_dm['mass'] < 1.01 * f_dm['mass'].min()])
            self.snapshot.wrap()
        elif center == 'all':
            pynbody.analysis.halo.center(self.snapshot)
        elif center == 'none':
            self.snapshot['pos']-=(self.snapshot['pos'].max(axis=0) + self.snapshot['pos'].min(axis=0))/2.0
        else:
            raise ValueError("Unknown centering type")

    def _perform_smoothing(self):
        try:
            logger.info("Looking for cached smoothing data...")
            smooth = pickle.load(open(self._smooth_cache_filename, 'rb'))
            if len(smooth) == len(self.snapshot):
                self.snapshot[self._name_smooth_array] = smooth
            else:
                raise ValueError("Incorrect number of particles in cached smoothing data")
            logger.info("...success!")
        except:
            logger.info("Generating smoothing data - this can take a while but will be cached for future runs")
            self.snapshot[self._name_smooth_array] = pynbody.sph.smooth(self.snapshot)
            try:
                pickle.dump(self.snapshot[self._name_smooth_array], open(self._smooth_cache_filename, 'wb'))
                logger.info("Smoothing data saved successfully")
            except IOError:
                logger.warning("Unable to save smoothing data to disk")


class TestDataLoader(AbstractDataLoader):
    def __init__(self, device: wgpu.GPUDevice, n_particles: int = config.TEST_DATA_NUM_PARTICLES_DEFAULT,
                 n_cells = 10, seed: int = 1337, with_cells = False, periodic = False):
        self._n_particles = n_particles
        self._gmm_weights = [0.5, 0.4, 0.1]  # should sum to 1
        self._gmm_means = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [6.0, 10.0, 0.0]])
        self._gmm_std = np.array([[20.0, 20.0, 20.0], [4.0, 0.2, 4.0], [2.0, 2.0, 3.0]])

        self._gmm_pos = self._generate_samples(seed)
        self._gmm_den = self._evaluate_density(self._gmm_pos)

        self._periodic = periodic

        if with_cells:
            self._cell_layout, ordering = cell_layout.CellLayout.from_positions(self._gmm_pos, self._gmm_pos.min()-1e-3,
                                                                                self._gmm_pos.max()+1, n_cells)
            self._gmm_pos = self._gmm_pos[ordering]
            self._gmm_den = self._gmm_den[ordering]

        super().__init__(device)

    def __len__(self):
        return self._n_particles

    def _evaluate_density(self, pos):
        # simple gaussian mixture model, density of particles per unit volume
        den = np.zeros(len(pos))
        for i in range(len(self._gmm_weights)):
            den += self._gmm_weights[i] * np.exp(
                -np.sum((pos - self._gmm_means[i]) ** 2 / self._gmm_std[i] ** 2, axis=1)) \
                   / ((2 * np.pi) ** 1.5 * np.prod(self._gmm_std[i]))
        return den * self._n_particles

    def _generate_samples(self, seed):
        # simple gaussian mixture model
        np.random.seed(seed)
        pos = np.empty((self._n_particles, 3), dtype=np.float32)
        if self._n_particles == 1:
            pos[0] = self._gmm_means[0]
        else:
            offset = 0
            for i in range(len(self._gmm_weights)):
                cpt_len = int(self._n_particles * self._gmm_weights[i])
                pos[offset:offset + cpt_len] = \
                    np.random.normal(size=(cpt_len, 3), scale=1.0).astype(np.float32) * self._gmm_std[np.newaxis, i,
                                                                                        :] + self._gmm_means[i]
                offset += cpt_len
            assert offset == self._n_particles
        return np.random.permutation(pos)

    def get_positions(self):
        return self._gmm_pos

    def get_smooth(self):
        sm = 2.0 / self._gmm_den ** 0.333333
        return sm

    def get_mass(self):
        return np.repeat(np.float32(1e-8), self._n_particles)

    def get_named_quantity(self, name):
        if name == "test-quantity":
            return np.sin(self._gmm_pos[:, 0]) * np.cos(self._gmm_pos[:, 1]) * np.cos(self._gmm_pos[:, 2]) * 1e-4
        else:
            raise KeyError("Unknown quantity name")

    def get_quantity_names(self):
        return ["test-quantity"]

    def get_quantity_label(self, quantity_name):
        if quantity_name is None:
            return r"test density / $M_{\odot} / \mathrm{kpc}^2$"
        elif quantity_name == "test-quantity":
            return "test quantity"
        else:
            return "unknown"

    def get_filename(self):
        return "test data"

    def get_periodicity_scale(self):
        return 100.0 if self._periodic else None

    def get_rgb_masses(self):
        rgb = np.empty((len(self._gmm_pos), 3), dtype=np.float32)
        rgb[:, 0] = abs(np.sin(self._gmm_pos[:, 0] / 10.0))
        rgb[:, 1] = abs(np.cos(self._gmm_pos[:, 1] / 10.0))
        rgb[:, 2] = abs(np.cos(self._gmm_pos[:, 2] / 10.0))
        return rgb



from __future__ import annotations

import time

from PySide6 import QtWidgets, QtCore

from ...recorder import VisualizationRecorder

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...visualizer import Visualizer


class RecordingSettingsDialog(QtWidgets.QDialog):
    def __init__(self, *args):
        super().__init__(*args)
        self.setWindowTitle("Recording settings")
        self._layout = QtWidgets.QVBoxLayout()
        self.setLayout(self._layout)

        # checkbox for smoothing:
        self._smooth_checkbox = QtWidgets.QCheckBox("Smooth timestream camera movements")
        self._smooth_checkbox.setChecked(True)
        self._layout.addWidget(self._smooth_checkbox)

        # leave some space:
        self._layout.addSpacing(10)

        # checkbox for including vmin/vmax:
        self._vmin_vmax_checkbox = QtWidgets.QCheckBox("Set vmin/vmax from timestream")
        self._vmin_vmax_checkbox.setChecked(True)
        self._layout.addWidget(self._vmin_vmax_checkbox)

        # checkbox for changing quantity:
        self._quantity_checkbox = QtWidgets.QCheckBox("Set quantity from timestream")
        self._quantity_checkbox.setChecked(True)
        self._layout.addWidget(self._quantity_checkbox)

        self._layout.addSpacing(10)

        # checkbox for showing colorbar:
        self._colorbar_checkbox = QtWidgets.QCheckBox("Show colorbar")
        self._colorbar_checkbox.setChecked(True)
        self._layout.addWidget(self._colorbar_checkbox)

        # checkbox for showing scalebar:
        self._scalebar_checkbox = QtWidgets.QCheckBox("Show scalebar")
        self._scalebar_checkbox.setChecked(True)
        self._layout.addWidget(self._scalebar_checkbox)

        self._layout.addSpacing(10)


        # select resolution from dropdown, with options half HD, HD, 4K
        self._resolution_dropdown = QtWidgets.QComboBox()
        self._resolution_dropdown.addItems(["Half HD (960x540)", "HD (1920x1080)", "4K (3840x2160)"])
        self._resolution_dropdown.setCurrentIndex(1)

        # select fps from dropdown, with options 24, 30, 60
        self._fps_dropdown = QtWidgets.QComboBox()
        self._fps_dropdown.addItems(["24 fps", "30 fps", "60 fps"])
        self._fps_dropdown.setCurrentIndex(1)

        # put resolution/fps next to each other horizontally:
        self._resolution_fps_layout = QtWidgets.QHBoxLayout()
        self._resolution_fps_layout.addWidget(self._resolution_dropdown)
        self._resolution_fps_layout.addWidget(self._fps_dropdown)
        self._layout.addLayout(self._resolution_fps_layout)

        self._layout.addSpacing(10)

        # cancel and save.. buttons:
        self._cancel_save_layout = QtWidgets.QHBoxLayout()
        self._cancel_button = QtWidgets.QPushButton("Cancel")
        self._cancel_button.clicked.connect(self.reject)
        self._save_button = QtWidgets.QPushButton("Save")
        # save button should be default:
        self._save_button.setDefault(True)
        self._save_button.clicked.connect(self.accept)
        self._cancel_save_layout.addWidget(self._cancel_button)
        self._cancel_save_layout.addWidget(self._save_button)
        self._layout.addLayout(self._cancel_save_layout)

        # show as a sheet on macos:
        #self.setWindowModality(QtCore.Qt.WindowModality.WindowModal)
        self.setWindowFlags(QtCore.Qt.WindowType.Sheet)

    @property
    def fps(self):
        return float(self._fps_dropdown.currentText().split()[0])

    @property
    def resolution(self):
        import re
        # use regexp
        # e.g. the string 'blah (123x456)' should map to tuple (123,456)
        match = re.match(r".*\((\d+)x(\d+)\)", self._resolution_dropdown.currentText())
        return int(match.group(1)), int(match.group(2))

    @property
    def smooth(self):
        return self._smooth_checkbox.isChecked()

    @property
    def set_vmin_vmax(self):
        return self._vmin_vmax_checkbox.isChecked()

    @property
    def set_quantity(self):
        return self._quantity_checkbox.isChecked()

    @property
    def show_colorbar(self):
        return self._colorbar_checkbox.isChecked()

    @property
    def show_scalebar(self):
        return self._scalebar_checkbox.isChecked()


class VisualizationRecorderWithQtProgressbar(VisualizationRecorder):

    def __init__(self, visualizer: Visualizer, parent_widget: QtWidgets.QWidget):
        super().__init__(visualizer)
        self._parent_widget = parent_widget

    def _progress_iterator(self, ntot):
        progress_bar = QtWidgets.QProgressDialog("Rendering to mp4...", "Stop", 0, ntot, self._parent_widget)
        progress_bar.setWindowModality(QtCore.Qt.WindowModality.WindowModal)
        progress_bar.forceShow()

        last_update = 0

        loop = QtCore.QEventLoop()

        try:
            for i in range(ntot):
                # updating the progress bar triggers a render in the main window, which
                # in turn is quite slow (because it can trigger software rendering
                # of resizable elements like the colorbar). So only update every half second or so.
                if time.time() - last_update > 0.5:
                    last_update = time.time()
                    progress_bar.setValue(i)

                    with self._visualizer.prevent_sph_rendering():
                        loop.processEvents(QtCore.QEventLoop.ProcessEventsFlag.AllEvents)

                    if progress_bar.wasCanceled():
                        break
                yield i

        finally:
            progress_bar.close()



import time 
import ipywidgets as widgets
from playwright.sync_api import Page
import pytest 

from IPython.display import display


from typing import Callable 

import topsy, topsy.canvas.jupyter


def poll_until_true(assertion: Callable, timeout=2, iteration_delay=0.01):
    start = time.time()
    while time.time() - start < timeout:
        if assertion():
            return True
        time.sleep(iteration_delay)
    return False

@pytest.fixture
def jupyter_vis(solara_test):
    vis = topsy.test(100, canvas_class = topsy.canvas.jupyter.VisualizerCanvas)
    display(vis)
    return vis


def test_colormap_name_select(jupyter_vis, page_session: Page):
    vis = jupyter_vis

    assert vis.colormap_name == "twilight_shifted"

    sel = page_session.locator("select:has-text('twilight_shifted')")
    sel.wait_for()
    sel.select_option("twilight")

    assert poll_until_true(lambda: vis.colormap_name == "twilight")

    assert vis.quantity_name == None

def test_quantity_name_select(jupyter_vis, page_session: Page):
    sel = page_session.locator(f"select:has-text('{topsy.config.PROJECTED_DENSITY_NAME}')")
    cb = page_session.locator("input[type='checkbox']")

    sel.wait_for()
    cb.wait_for()

    assert cb.is_checked()
    sel.select_option("test-quantity")


    assert poll_until_true(lambda: jupyter_vis.quantity_name == "test-quantity")

    # check that log quantity is no longer selected    
    assert poll_until_true(lambda: not cb.is_checked())

def test_alter_range(jupyter_vis, page_session: Page):
    vis = jupyter_vis

    min_slider = page_session.locator("div.noUi-handle-lower")
    max_slider = page_session.locator("div.noUi-handle-upper")

    min_slider.wait_for()
    max_slider.wait_for()

    vmin_orig = vis.vmin
    vmax_orig = vis.vmax

    # Use keyboard to move the slider instead of drag_to to avoid pointer event interception issues
    min_slider.focus()
    page_session.keyboard.press("ArrowLeft")
    page_session.keyboard.press("ArrowLeft")
    max_slider.focus()
    page_session.keyboard.press("ArrowRight")
    page_session.keyboard.press("ArrowRight")

    assert poll_until_true(lambda: vis.vmin < vmin_orig)
    assert poll_until_true(lambda: vis.vmax > vmax_orig)

def test_rgb_map(solara_test, page_session: Page):
    vis = topsy.test(100, canvas_class = topsy.canvas.jupyter.VisualizerCanvas, rgb=True)
    display(vis)

    # at the moment we just check this actually gives the alternative panel
    sel = page_session.locator("text=gamma")
    sel.wait_for()

    


from __future__ import annotations

import copy
import numpy as np
import wgpu
import pynbody

from logging import getLogger

from .util import load_shader, preprocess_shader, TimeGpuOperation
from .drawreason import DrawReason

from . import config, performance

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .visualizer import Visualizer

logger = getLogger(__name__)


class SPH:
    render_format = wgpu.TextureFormat.rg32float
    _nchannels_input = 2
    _nchannels_output = 2
    _output_dtype = np.float32

    def __init__(self, visualizer: Visualizer, render_resolution,
                 wrapping = False, share_render_progression=None):
        self._visualizer = visualizer

        logger.info(f"Creating SPH renderer with resolution {render_resolution}")
        self._render_texture = visualizer.device.create_texture(
            size=(render_resolution, render_resolution, 1),
            usage=wgpu.TextureUsage.RENDER_ATTACHMENT |
                  wgpu.TextureUsage.TEXTURE_BINDING |
                  wgpu.TextureUsage.COPY_SRC,
            format=self.render_format,
            label="sph_render_texture",
        )

        self._render_resolution = render_resolution

        self._device : wgpu.GPUDevice = visualizer.device
        self._wrapping = wrapping
        self._kernel = None

        self._render_timer = TimeGpuOperation(self._device)

        if share_render_progression is not None:
            self._render_progression = share_render_progression
        else:
            self._render_progression = self._visualizer.data_loader.get_render_progression()

        self._setup_shader_module()
        self._setup_transform_buffer()
        self._setup_kernel_texture()
        self._setup_render_pipeline()

        self.scale = config.DEFAULT_SCALE
        self.min_pixels = 0.0    # minimum size of softening, in pixels, to qualify for rendering
        self.max_pixels = np.inf # maximum size of softening, in pixels, to qualify for rendering
        self.rotation_matrix = np.eye(3)
        self.position_offset = np.zeros(3)

    def _get_depth_renderer(self) -> SPH:
        """Returns a SPH renderer that will generate the depth in the scene"""
        renderer = DepthSPH(self._visualizer, self._render_resolution, wrapping=self._wrapping,
                   share_render_progression=copy.copy(self._render_progression))
        renderer.rotation_matrix = self.rotation_matrix
        renderer.position_offset = self.position_offset
        renderer.scale = self.scale
        return renderer


    def get_depth_image(self, depth_renderer_reason=DrawReason.CHANGE) -> np.ndarray:
        """Produces and returns the weighted depth image in the scene, used for finding points of interest in the UI

        A renderer reason may be passed to force different quality settings. For most purposes, a rough (real-time)
        render is sufficient, so DrawReason.CHANGE is a good choice. However, real-time rendering on test machines
        may be slow, which leads DrawReason.CHANGE renders to use only a small fraction of particles. Therefore, for
        testing purposes DrawReason.EXPORT may be required.
        """
        depth_renderer = self._get_depth_renderer()
        depth_renderer.render(depth_renderer_reason) # CHANGE should normally be good enough; EXPORT for reliability
        depth_viewport = depth_renderer.get_image(averaging=True)

        # transform from viewport to simulation units
        return (depth_viewport - 0.5)*self.scale*2.0

    def get_image(self, averaging) -> np.ndarray:
        """Reads and returns the last rendered SPH image.

        Note that this call does not actually trigger a render. Call render() first to generate the image."""
        nchannels = self._nchannels_output
        np_dtype = self._output_dtype
        bytes_per_pixel = nchannels * np.dtype(np_dtype).itemsize

        im = self._device.queue.read_texture({'texture': self.get_output_texture(), 'origin': (0, 0, 0)},
                                            {'bytes_per_row': bytes_per_pixel * self._render_resolution},
                                            (self._render_resolution, self._render_resolution, 1))
        np_im = np.frombuffer(im, dtype=np_dtype).reshape((self._render_resolution, self._render_resolution, nchannels))

        if nchannels == 4:
            if averaging:
                raise ValueError("Averaging not supported for RGBA output images")
            np_im = np_im[:, :, :3] * self.last_render_mass_scale
        elif averaging:
            np_im = np_im[:, :, 1] / np_im[:, :, 0]
        else:
            np_im = np_im[:, :, 0] * self.last_render_mass_scale

        return np_im


    def get_output_texture(self) -> wgpu.Texture:
        return self._render_texture


    def _setup_shader_module(self, flags=None):
        if flags is None:
            flags = []
        wrap_flag = "WRAPPING" if self._wrapping else "NO_WRAPPING"
        type_flag = "CHANNELED" if self._nchannels_input == 3 else "WEIGHTED"
        flags += [wrap_flag, type_flag]
        code = preprocess_shader(load_shader("sph.wgsl"), flags)

        self._shader = self._device.create_shader_module(code=code, label="sph")

    def _setup_transform_buffer(self):
        self._transform_buffer = self._device.create_buffer(
            size=(4*4*4) + 4 + 8 + 8 + 12,  # 4x4 float32 matrix + one float32 scale + float32 min,max size + int32 x 2 + padding
            usage=wgpu.BufferUsage.UNIFORM | wgpu.BufferUsage.COPY_DST
        )

    def _setup_render_pipeline(self):
        self._bind_group_layout = \
            self._device.create_bind_group_layout(
                label="sph_bind_group_layout",
                entries=[
                    {
                        "binding": 0,
                        "visibility": wgpu.ShaderStage.VERTEX,
                        "buffer": {"type": wgpu.BufferBindingType.uniform},
                    },
                    {
                        "binding": 1,
                        "visibility": wgpu.ShaderStage.FRAGMENT,
                        "texture": {"sample_type": wgpu.TextureSampleType.float,
                                    "view_dimension": wgpu.TextureViewDimension.d2},
                    },
                    {
                        "binding": 2,
                        "visibility": wgpu.ShaderStage.FRAGMENT,
                        "sampler": {"type": wgpu.SamplerBindingType.filtering},
                    },
                ]
            )

        self._bind_group = \
            self._device.create_bind_group(
                label="sph_bind_group",
                layout=self._bind_group_layout,
                entries=[
                    {"binding": 0,
                     "resource": {
                         "buffer": self._transform_buffer,
                         "offset": 0,
                         "size": self._transform_buffer.size
                      }
                     },
                    {"binding": 1,
                        "resource": self._kernel_texture.create_view(),
                    },
                    {"binding": 2,
                        "resource": self._kernel_sampler,
                    }
                ]
            )

        self._pipeline_layout = \
            self._device.create_pipeline_layout(
                label="sph_pipeline_layout",
                bind_group_layouts=[self._bind_group_layout]
            )

        vertex_format = wgpu.VertexFormat.float32x3

        if self._nchannels_input == 3 :
            channel_buffers = [{
                "array_stride": 12,
                "step_mode": wgpu.VertexStepMode.instance,
                "attributes": [
                    {
                        "format": vertex_format,
                        "offset": 0,
                        "shader_location": 1,
                    }
                ]
            } ]
        else:
            channel_buffers = [ {
                                "array_stride": 4,
                                "step_mode": wgpu.VertexStepMode.instance,
                                "attributes": [
                                    {
                                        "format": wgpu.VertexFormat.float32,
                                        "offset": 0,
                                        "shader_location": i+1,
                                    }
                                ]
                            } for i in range(self._nchannels_input)]

        self._render_pipeline = \
            self._device.create_render_pipeline(
                layout=self._pipeline_layout,
                label="sph_render_pipeline",
                vertex = {
                    "module": self._shader,
                    "entry_point": "vertex_main",
                    "buffers": [
                        {
                            "array_stride": 16,
                            "step_mode": wgpu.VertexStepMode.instance,
                            "attributes": [
                                {
                                    "format": wgpu.VertexFormat.float32x4,
                                    "offset": 0,
                                    "shader_location": 0,
                                }
                            ]
                        },


                    ] + channel_buffers
                },
                primitive={
                    "topology": wgpu.PrimitiveTopology.triangle_list,
                },
                depth_stencil=None,
                multisample=None,
                fragment={
                    "module": self._shader,
                    "entry_point": "fragment_main",
                    "targets": [
                        {
                            "format": self.render_format,
                            "blend": {
                                "color": {
                                    "src_factor": wgpu.BlendFactor.one,
                                    "dst_factor": wgpu.BlendFactor.one,
                                    "operation": wgpu.BlendOperation.add,
                                },
                                "alpha": {
                                    "src_factor": wgpu.BlendFactor.one,
                                    "dst_factor": wgpu.BlendFactor.one,
                                    "operation": wgpu.BlendOperation.add,
                                },
                              }
                        }
                    ]
                }
            )

    def _update_transform_buffer(self):
        model_displace = np.array([[1.0, 0, 0, self.position_offset[0]],
                                   [0, 1.0, 0, self.position_offset[1]],
                                   [0, 0, 1.0, self.position_offset[2]],
                                   [0, 0, 0.0, 1.0]])

        # self._transform is the transformation around the origin (fine for opengl)
        # but in webgpu, the clip space in the z direction is [0,1]
        # so we need a matrix that brings z=0. to z=0.5 and squishes the z direction
        # so that the clipping is the same in all dimensions
        clipcoord_displace = np.array([[1.0, 0, 0, 0.0],
                             [0, 1.0, 0, 0.0],
                             [0, 0, 0.5, 0.5],
                             [0, 0, 0.0, 1.0]])
        transform = np.zeros((4, 4))

        transform[:3,:3] = self.rotation_matrix
        rotation_and_scaling = transform / self.scale

        rotation_and_scaling[3, 3] = 1.0  # w should be unchanged after transform

        scaled_displaced_transform = (clipcoord_displace @ rotation_and_scaling @ model_displace).T
        transform_params_dtype = [("transform", np.float32, (4, 4)),
                                  ("scale_factor", np.float32, (1,)),
                                  ("min_max_size", np.float32, (2,)),
                                  ("boxsize_by_2_clipspace", np.float32, (1,)),
                                  ("padding", np.int32, (1,))]
        transform_params = np.zeros((), dtype=transform_params_dtype)
        transform_params["transform"] = scaled_displaced_transform
        transform_params["scale_factor"] = 1. / self.scale
        if self._visualizer.periodicity_scale is not None:
            transform_params["boxsize_by_2_clipspace"] = 0.5 * \
                                                         self._visualizer.periodicity_scale / self.scale
        else:
            transform_params["boxsize_by_2_clipspace"] = 0.0

        resolution = self._render_texture.width
        assert resolution == self._render_texture.height

        # min_max_size to be sent in viewport coordinates
        transform_params["min_max_size"] = (2.*self.min_pixels/resolution, 2.*self.max_pixels/resolution)

        self.last_transform_params = transform_params

        self._device.queue.write_buffer(self._transform_buffer, 0, transform_params)



    def render(self, draw_reason=DrawReason.CHANGE):
        performance.signposter.emit_event("Start SPH render")

        if draw_reason == DrawReason.PRESENTATION_CHANGE:
            return

        if draw_reason != DrawReason.REFINE:
            self._render_progression.select_sphere(-self.position_offset, self.scale*1.2)
            self._update_transform_buffer()

        clear = self._render_progression.start_frame(draw_reason)

        while block := self._render_progression.get_block(self._render_timer.total_time_in_frame()):
            encoded_render_pass = self.encode_render_pass(clear=clear)
            self._visualizer.particle_buffers.update_particle_ranges(*block)
            with self._render_timer:
                # we only time this part, because otherwise the timing is very unstable in interactive
                # use where most of the time we are not updating particle ranges
                self._device.queue.submit([encoded_render_pass])
            self._render_progression.end_block(self._render_timer.total_time_in_frame())
            clear = False

        self._render_timer.end_frame()

        self.last_render_mass_scale = self._render_progression.end_frame_get_scalefactor()
        self.last_render_fps = 1.0 / self._render_timer.running_mean_duration

    def needs_refine(self):
        return self._render_progression.needs_refine()

    def encode_render_pass(self, clear=True) -> wgpu.GPUCommandBuffer:
        command_encoder: wgpu.GPUCommandEncoder = self._device.create_command_encoder(label='sph_render')
        view: wgpu.GPUTextureView = self._render_texture.create_view()
        sph_render_pass: wgpu.GPURenderPassEncoder = command_encoder.begin_render_pass(
            color_attachments=[
                {
                    "view": view,
                    "resolve_target": None,
                    "clear_value": (0.0, 0.0, 0.0, 0.0),
                    "load_op": wgpu.LoadOp.clear if clear else wgpu.LoadOp.load,
                    "store_op": wgpu.StoreOp.store,
                }
            ]
        )
        sph_render_pass.set_pipeline(self._render_pipeline)

        vb_assignment = ['pos_smooth']
        if self._nchannels_input == 2:
            vb_assignment += ['mass', 'quantity']
        elif self._nchannels_input == 3:
            vb_assignment += ['rgb_masses']
        else:
            raise ValueError("Unexpected number of channels")

        self._visualizer.particle_buffers.specify_vertex_buffer_assignment(vb_assignment)
        sph_render_pass.set_bind_group(0, self._bind_group, [],
                                       0, 99)

        self._visualizer.particle_buffers.issue_draw_indirect(sph_render_pass)
        sph_render_pass.end()

        return command_encoder.finish()

    def _get_kernel_at_resolution(self, n_samples):
        if self._kernel is None:
            try:
                self._kernel = pynbody.sph.Kernel2D()
            except AttributeError:
                # pynbody v2:
                self._kernel = pynbody.sph.kernels.Kernel2D()

        # sph kernel is sampled at the centre of the pixels, and the full grid ranges from -2 to 2.
        # thus the left hand most pixel is at -2+2/n_samples, and the right hand most pixel is at 2-2/n_samples.
        pixel_centres = np.linspace(-2+2./n_samples, 2-2./n_samples, n_samples)
        x, y = np.meshgrid(pixel_centres, pixel_centres)
        distance = np.sqrt(x ** 2 + y ** 2)

        # The below could easily be optimized but doesn't seem worth it
        kernel_im = np.array([self._kernel.get_value(d) for d in distance.flatten()]).reshape(n_samples, n_samples)

        # make kernel explicitly mass conserving; naive pixelization makes it not automatically do this.
        # It should be normalized such that the integral over the kernel is 1/h^2. We have h=1 here, and the
        # full width is 4h, so the width of a pixel is dx=4/n_samples. So we need to multiply by dx^2=(n_samples/4)^2.
        # This results in a correction of a few percent, typically; not huge but not negligible either.
        #
        # (Obviously h!=1 generally, so the h^2 normalization occurs within the shader later.)
        kernel_im *= (n_samples / 4) ** 2 / kernel_im.sum()

        return kernel_im

    def _setup_kernel_texture(self, n_samples=64, n_mip_levels = 4):
        if hasattr(SPH, "_kernel_texture"):
            # As things stand, the kernel texture is actually shared between instances of SPH for efficiency.
            return

        SPH._kernel_texture = self._device.create_texture(
            label="kernel_texture",
            size=(n_samples, n_samples, 1),
            dimension=wgpu.TextureDimension.d2,
            format=wgpu.TextureFormat.r32float,
            mip_level_count=n_mip_levels,
            sample_count=1,
            usage=wgpu.TextureUsage.COPY_DST | wgpu.TextureUsage.TEXTURE_BINDING,
        )

        for i in range(0, n_mip_levels):
            self._device.queue.write_texture(
                {
                    "texture": self._kernel_texture,
                    "mip_level": i,
                    "origin": (0, 0, 0),
                },
                self._get_kernel_at_resolution(n_samples//2**i).astype(np.float32).tobytes(),
                {
                    "offset": 0,
                    "bytes_per_row": 4 * n_samples // 2**i,
                },
                (n_samples//2**i, n_samples//2**i, 1)
            )


        SPH._kernel_sampler = self._device.create_sampler(label="kernel_sampler",
                                                           mag_filter=wgpu.FilterMode.linear, )

class RGBSPH(SPH):
    render_format = wgpu.TextureFormat.rgba32float
    _nchannels_input = 3
    _nchannels_output = 4
    _output_dtype = np.float32

class DepthSPH(SPH):
    """Renders a map of the depth of the particles in the scene."""
    def _setup_shader_module(self, flags=None):
        if flags is None:
            flags = []
        super()._setup_shader_module(flags+["DEPTH"])



"""Tools for measuring performance"""

try:
    from os_signpost import Signposter
    signposter = Signposter("com.pynbody.topsy", Signposter.Category.PointsOfInterest)
except ImportError:
    # Most of the time we won't have this module, so make a dummy
    class DummySignposter:
        def __init__(self):
            pass

        def begin_interval(self, *args, **kwargs):
            pass

        def emit_event(self, *args, **kwargs):
            pass

        def use_interval(*args, **kwds):
            pass

    signposter = DummySignposter()


import numpy as np
import pytest

from topsy import progressive_render, config
from topsy.drawreason import DrawReason

def _get_single_block(blocks):
    """Helper function to convert from list of blocks to single block"""
    assert len(blocks)==2
    assert len(blocks[0])==1
    assert len(blocks[1])==1
    return blocks[0][0], blocks[1][0]

def test_initial_recommendations():
    # Test the initial recommendation for a small number of particles
    render_progression = progressive_render.RenderProgression(config.INITIAL_PARTICLES_TO_RENDER//2)
    render_progression.start_frame(DrawReason.INITIAL_UPDATE)
    assert _get_single_block(render_progression.get_block(0.0)) == (0, config.INITIAL_PARTICLES_TO_RENDER//2)

    # Test the initial recommendation for a large number of particles
    render_progression = progressive_render.RenderProgression(config.INITIAL_PARTICLES_TO_RENDER*2)
    render_progression.start_frame(DrawReason.INITIAL_UPDATE)
    assert _get_single_block(render_progression.get_block(0.0)) == (0, config.INITIAL_PARTICLES_TO_RENDER)

def test_export_recommendations():
    render_progression = progressive_render.RenderProgression(config.INITIAL_PARTICLES_TO_RENDER * 2)
    render_progression.start_frame(DrawReason.EXPORT)
    assert _get_single_block(render_progression.get_block(0.0)) == (0, config.INITIAL_PARTICLES_TO_RENDER*2)
    render_progression.end_block(0.1)
    assert render_progression.get_block(1.0) is None

def test_progression():
    # Test the progression of recommendations
    render_progression = progressive_render.RenderProgression(1000, 100)
    render_progression.start_frame(DrawReason.CHANGE)

    # Simulate rendering a block of particles
    start_index, num_to_render = _get_single_block(render_progression.get_block(0.0))
    assert start_index == 0
    assert num_to_render == 100

    # Simulate ending the block and reporting time taken
    render_progression.end_block(0.5/config.TARGET_FPS)

    # Check the next recommendation. We're half way through.
    start_index, num_to_render = _get_single_block(render_progression.get_block(0.5/config.TARGET_FPS))
    assert start_index == 100
    assert num_to_render == 50 # hasn't updated the expected rendering number, just looks at time remaining

    render_progression.end_block(1./config.TARGET_FPS)

    # Check the end of the frame
    assert render_progression.get_block(1./config.TARGET_FPS) is None

    # We've rendered 150 particles.
    assert render_progression.end_frame_get_scalefactor() == 1000./150


def test_timeout_and_progression():
    # Test the timeout and progression of recommendations
    render_progression = progressive_render.RenderProgression(1000, 100)
    render_progression.start_frame(DrawReason.CHANGE)

    # Simulate a long time elapsed
    block = _get_single_block(render_progression.get_block(0.0))
    assert block is not None

    render_progression.end_block(1.0)  # far too long!
    # Check that the next recommendation is None
    block = render_progression.get_block(1.0)
    assert block is None

    sf = render_progression.end_frame_get_scalefactor()
    assert sf == 10.0

    assert render_progression.needs_refine()

    render_progression.start_frame(DrawReason.REFINE)
    start, num = _get_single_block(render_progression.get_block(0.0))
    assert start == 100
    assert num == int(100/config.TARGET_FPS) # took one second to render 100 particles, so recommendation should be this


def test_always_one_block():
    render_progression = progressive_render.RenderProgression(1000, 100)
    render_progression.start_frame(DrawReason.CHANGE)

    block = _get_single_block(render_progression.get_block(1.0)) # simulate a long time elapsed, but should still render at least one block
    assert block is not None

def test_no_render_on_presentation_change():
    render_progression = progressive_render.RenderProgression(1000, 100)
    render_progression.start_frame(DrawReason.CHANGE)

    # Simulate rendering the first frame
    t = 0.0
    while (block := render_progression.get_block(t)) is not None:
        t+=1e-5
        render_progression.end_block(t)

    render_progression.end_frame_get_scalefactor()
    assert not render_progression.needs_refine()

    render_progression.start_frame(DrawReason.PRESENTATION_CHANGE)

    block = render_progression.get_block(0.0)
    assert block is None

    render_progression.end_frame_get_scalefactor()
    assert not render_progression.needs_refine()

def test_no_frame_exception():
    render_progression = progressive_render.RenderProgression(1000, 100)
    with pytest.raises(RuntimeError):
        _get_single_block(render_progression.get_block(0.0))

def test_export():
    """Test that export always recommends the full resolution"""

    render_progression = progressive_render.RenderProgression(1000, 100)
    render_progression.start_frame(DrawReason.EXPORT)
    block = _get_single_block(render_progression.get_block(0.0))
    assert block == (0, 1000)

def test_always_one_particle():
    """Test that always at least one particle is recommended (even if it takes a long time)"""

    render_progression = progressive_render.RenderProgression(1000, 3)
    render_progression.start_frame(DrawReason.CHANGE)

    # Simulate a long time elapsed
    block = _get_single_block(render_progression.get_block(0.0))
    assert block is not None

    render_progression.end_block(1.0)  # far too long!

    assert render_progression.get_block(1.0) is None # end of this frame
    render_progression.end_frame_get_scalefactor()
    assert render_progression.needs_refine()

    # Check that the refinement consists of at least one particle, even though technically
    # we would have rounded this down to zero particles
    render_progression.start_frame(DrawReason.REFINE)
    block = _get_single_block(render_progression.get_block(1.0))
    assert block == (3,1)

@pytest.fixture
def cell_progressive_render(num_particles=100000, num_side=10, num_part_to_take = 100):
    np.random.seed(1337)
    pos = np.random.uniform(0.0, 1.0, (num_particles, 3))

    cell_layout, order = progressive_render.CellLayout.from_positions(pos, 0.0, 1.0, num_side)
    pos = pos[order]

    render_progression = progressive_render.RenderProgressionWithCells(cell_layout, len(pos), num_part_to_take)

    return render_progression, pos

def test_blocks_with_layout(cell_progressive_render):

    render_progression, pos = cell_progressive_render
    cell_layout = render_progression._cell_layout

    total_particles = 0

    rendered = np.zeros(len(pos), dtype=np.int32)

    render_progression.start_frame(DrawReason.CHANGE)
    first_render = True

    while True:

        block = render_progression.get_block(0.0)

        for start, length in zip(*block):
            # find which cell this block belongs to
            cell_index = cell_layout.cell_index_from_offset(start)
            assert length!=0 # should not return zero length blocks
            cell_index_at_end = cell_layout.cell_index_from_offset(start+length-1)
            assert cell_index == cell_index_at_end
            total_particles+=length
            rendered[start:start+length]+=1

        if first_render:
            assert total_particles > 95 and total_particles < 105

        render_progression.end_block(0.0001)
        render_progression.end_frame_get_scalefactor()

        if render_progression.needs_refine():
            first_render = False
            render_progression.start_frame(DrawReason.REFINE)
        else:
            break

    assert (rendered==1).all()

    # check that render_progression returns None at end of frame

    render_progression.start_frame(DrawReason.CHANGE)
    npart = 0
    while (block := render_progression.get_block(0.0)):
        starts, lens = block
        npart+=lens.sum()
        render_progression.end_block(0.0)
    assert npart == len(pos)

def test_spatial_limits(cell_progressive_render):
    render_progression, pos = cell_progressive_render

    render_progression.select_sphere((0.5, 0.5, 0.5), 0.1)

    render_progression.start_frame(DrawReason.CHANGE)
    rendered = np.zeros(len(pos), dtype=np.int32)
    while (block:=render_progression.get_block(0.0)):
        for start, length in zip(*block):
            rendered[start:start+length]+=1

        render_progression.end_block(0.0)

    assert rendered.max() == 1

    r = np.linalg.norm(pos-0.5, axis=1)
    rendered_r = r[rendered==1]
    unrendered_r = r[rendered==0]

    assert (rendered_r<0.4).all()
    assert (unrendered_r>0.1).all()

def test_export_very_large():
    num_renders = 5
    render_progression = progressive_render.RenderProgression(config.MAX_PARTICLES_PER_EXPORT_RENDERCALL * num_renders)
    render_progression.start_frame(DrawReason.EXPORT)

    for blocknum in range(num_renders):
        # pretend we have a crazy long-running render. We should render the whole thing, but in a series
        # of blocks
        block = render_progression.get_block(100.0*blocknum)
        assert block is not None
        assert block[0][0] == config.MAX_PARTICLES_PER_EXPORT_RENDERCALL * blocknum
        assert block[1][0] == config.MAX_PARTICLES_PER_EXPORT_RENDERCALL
        render_progression.end_block(100.0 * (blocknum + 1))

    assert render_progression.get_block(100.0*num_renders) is None # finished now!

    clear = render_progression.start_frame(DrawReason.EXPORT)

    assert clear




import numpy as np

from topsy.recorder import interpolator

def test_step_interpolator():
    timestream = [(0.4, None), (1.0, "Hello"), (2.0, "World")]
    interp = interpolator.StepInterpolator(timestream)
    assert interp(0.0) is interp.no_value
    assert interp(0.5) is None
    assert interp(1.0) == "Hello"
    assert interp(1.5) is interp.no_value
    assert interp(1.9) is interp.no_value
    assert interp(2.1) == "World"

def test_linear_interpolator():
    timestream = [(0.0, 0.0), (1.0, 1.0), (2.0, 4.0), (4.0, 0.0)]
    interp = interpolator.LinearInterpolator(timestream)
    assert interp(0.0) == 0.0
    assert interp(0.5) == 0.5
    assert interp(1.0) == 1.0
    assert interp(1.5) == 2.5
    assert interp(2.0) == 4.0
    assert interp(3.0) == 2.0
    assert interp(4.0) == 0.0
    assert interp(4.01) is interp.no_value

def test_rotation_interpolator():
    timestream = [(0.0, np.eye(3)), (1.0, np.array([[0.0, 1.0, 0.0], [-1.0, 0.0, 0.0], [0.0, 0.0, 1.0]]))]
    interp = interpolator.RotationInterpolator(timestream)
    assert np.allclose(interp(0.0), np.eye(3))
    midway = interp(0.5)
    assert np.allclose(midway @ midway.T, np.eye(3))
    assert midway[0,0]<1.0 and midway[0,0]>0.0
    assert midway[0,1]<1.0 and midway[0,1]>0.0
    assert np.allclose(interp(1.0), np.array([[0.0, 1.0, 0.0], [-1.0, 0.0, 0.0], [0.0, 0.0, 1.0]]))

def test_smoothed_linear_interpolator():
    timestream = [(0.0, 0.0), (1.0, 1.0), (2.0, 4.0), (4.0, 0.0)]
    interp = interpolator.SmoothedLinearInterpolator(timestream, smoothing=0.5)
    assert np.allclose(interp(0.0), 0.18728488638447055)
    assert np.allclose(interp(0.5), 0.5833226799824336)
    assert np.allclose(interp(1.0), 1.3206482380039408)
    assert np.allclose(interp(1.5), 2.3157963036465143)
    assert np.allclose(interp(3.9), 0.5695905129936958)
    assert np.allclose(interp(4.0), 0.4616432412285651)
    assert interp(4.1) is interp.no_value
    # check smoothness
    assert abs(np.diff(np.diff([interp(x) for x in np.arange(0.0,4.0,0.05)]))).max()<0.02

def test_smoothed_rotation_interpolator():
    timestream = [(0.0, np.eye(3)), (1.0, np.array([[0.0, 1.0, 0.0], [-1.0, 0.0, 0.0], [0.0, 0.0, 1.0]]))]
    interp = interpolator.SmoothedRotationInterpolator(timestream, smoothing=0.5)
    for x in np.arange(0.0,1.0,0.1):
        assert np.allclose(interp(x) @ interp(x).T, np.eye(3))

def test_smoothed_step_interpolator():
    timestream = [(0.0, 0.0), (1.0, 5.0), (2.0, 0.0)]
    interp = interpolator.SmoothedStepInterpolator(timestream, smoothing=0.5)
    assert interp(0.1) == 0.0
    assert interp(0.5) is interp.no_value
    assert interp(1.0) == 0.0
    assert interp(1.125) == 1.25
    assert interp(1.25) == 2.5
    assert interp(1.5) == 5.0
    assert interp(1.75) is interp.no_value
    assert interp(1.99) is interp.no_value
    assert interp(2.0) == 5.0
    assert interp(2.25) == 2.5



from __future__ import annotations

from PySide6 import QtWidgets, QtCore


class MyLineEdit(QtWidgets.QLineEdit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._timer = QtCore.QTimer()
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self.selectAll)

    def focusInEvent(self, event):
        super().focusInEvent(event)
        self._timer.start(0)



import enum

class DrawReason(enum.Enum):
    """Enum to specify the reason for a draw (which may affect detailed behaviour)"""
    INITIAL_UPDATE = 1      # render from scratch
    CHANGE = 2              # a change has occurred, possibly from the UI
    REFINE = 3              # render the SPH at full resolution, within an interactive context
    PRESENTATION_CHANGE = 4 # i.e. don't rerender SPH
    EXPORT = 5              # full rendering, always at full resolution



from __future__ import annotations

from topsy.drawreason import DrawReason

from . import sph
from . import overlay

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .visualizer import Visualizer

import numpy as np
import wgpu

class PeriodicSPHAccumulationOverlay(overlay.Overlay):
    _blending = {
        "src_factor": wgpu.BlendFactor.one,
        "dst_factor": wgpu.BlendFactor.one,
        "operation": wgpu.BlendOperation.add
    }

    def __init__(self, visualizer: Visualizer, source_texture: wgpu.GPUTexture):
        self._texture = source_texture
        self.num_repetitions = 0
        self.panel_scale = 1.0
        self.rotation_matrix = np.eye(3)
        super().__init__(visualizer, source_texture.format)

    def _setup_texture(self):
        pass

    def get_clipspace_coordinates(self, width, height) -> tuple[float, float, float, float]:
        return -1.0, -1.0, 2.0, 2.0

    def get_instance_offsets_and_weights(self):
        offsets = []
        weights = []

        for xoff in range(-self.num_repetitions, self.num_repetitions + 1):
            for yoff in range(-self.num_repetitions, self.num_repetitions + 1):
                for zoff in range(-self.num_repetitions, self.num_repetitions + 1):
                    offset = self.rotation_matrix @ np.array([xoff,yoff,zoff], dtype=np.float32)
                    if abs(offset[2]) < 1.0:
                        offsets.append(offset[:2])
                        z = abs(offset[2])
                        # the weight be 1 for 0<z<0.5, and smoothly decrease to 0 over the range 0.5<z<1.0
                        if z>0.5:
                            weight = 1.0 - 2.0*(z-0.5)
                        else:
                            weight = 1.0
                        weights.append(weight)

        return np.array(offsets, dtype=np.float32) * self.panel_scale, np.array(weights, dtype=np.float32)

    def render_contents(self) -> np.ndarray:
        # must be implemented, but should never be called because texture is provided externally
        raise RuntimeError("SPHAccumulationOverlay.render_contents() should never be called")

class PeriodicSPH(sph.SPH):
    def __init__(self, visualizer, render_size):
        super().__init__(visualizer, render_size, wrapping=True)
        self._periodic_texture =  visualizer.device.create_texture(
                size=self._render_texture.size,
                format=self._render_texture.format,
                usage=self._render_texture.usage,
                label=f"proxy_sph"
            )
        self._accumulator = PeriodicSPHAccumulationOverlay(visualizer, self._render_texture)

    def get_output_texture(self) -> wgpu.Texture:
        return self._periodic_texture

    def render(self, draw_reason=DrawReason.CHANGE):
        if draw_reason == DrawReason.PRESENTATION_CHANGE:
            return

        super().render(draw_reason)

        command_encoder = self._visualizer.device.create_command_encoder(label="PeriodicSPH")
        self._accumulator.num_repetitions = 2
        self._accumulator.rotation_matrix = self.rotation_matrix
        self._accumulator.panel_scale = self._visualizer.periodicity_scale / self._visualizer.scale

        self._accumulator.encode_render_pass(command_encoder, self._periodic_texture.create_view(), True)

        encoded_render_pass = command_encoder.finish()
        self._device.queue.submit([encoded_render_pass])




from IPython.display import display
import ipywidgets as widgets
from typing import Callable, Any

from rendercanvas.jupyter import RenderCanvas, loop
from . import VisualizerCanvasBase
from .generic_colormap import ColorMapController, RGBMapController, ControlSpec, LayoutSpec
from ..config import JUPYTER_UI_LAG
from .. import colormap

class VisualizerCanvas(VisualizerCanvasBase, RenderCanvas):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._allow_events = True

    def request_draw(self, function=None):

        # As a side effect, wgpu gui layer stores our function call, to enable it to be
        # repainted later. But we want to distinguish such repaints and handle them
        # differently, so we need to replace the function with our own

        #def function_wrapper():
        #    function()
        #    self._subwidget.draw_frame = lambda: self._visualizer.draw(DrawReason.PRESENTATION_CHANGE)
        # TODO: above needs to be hacked for jupyter

        super().request_draw(function)



    @classmethod
    def call_later(cls, delay, fn, *args):
        loop.call_later(delay, fn, *args)

    def ipython_display_with_widgets(self):
        """Display the canvas in a Jupyter notebook with widgets."""
        color_controls = self.build_color_controls()

        # stack canvas, dropdown and range slider
        display(widgets.VBox([self, color_controls]))

    


    def build_color_controls(self) -> widgets.Widget:
        """
        Return a nested ipywidget (HBox/VBox) tree driven by the generic ColorMapController.get_layout() spec.
        """
        if isinstance(self._visualizer.colormap, colormap.RGBColormap):
            self._controller = RGBMapController(self._visualizer, self._refresh_ui)
        elif isinstance(self._visualizer.colormap, colormap.Colormap):
            self._controller = ColorMapController(self._visualizer, self._refresh_ui)
        else:
            self._controller = None

        if self._controller:
            self._controls = self.convert_layout_to_widget(self._controller.get_layout())
        else:
            self._controls = widgets.HTML("<b>No colormap controls available</b>")
        return self._controls

    def _callback(self, callback: Callable[[Any], None], value: Any):
        if not self._allow_events:
            return
        callback(value)
       
    
    def _refresh_ui(self):
        """Walk the layout and update all values, including slider ranges."""
        if not hasattr(self, "_controller"):
            return
        root_spec = self._controller.get_layout()
        self._allow_events = False
        try:
            self.update_widget(root_spec, self._controls)
        finally:
             # re-enable events after a delay, to allow the UI to settle (eugh! surely must be a better way?)
            self.call_later(JUPYTER_UI_LAG, lambda: setattr(self, "_allow_events", True))

    def convert_layout_to_widget(self, spec) -> widgets.Widget:
        children = []
        for child in spec.children:
            if isinstance(child, ControlSpec):
                children.append(self.make_widget(child))
            else:
                children.append(self.convert_layout_to_widget(child))
        if spec.type == "hbox":
            return widgets.HBox(children)
        else:
            return widgets.VBox(children)

    def make_widget(self, spec):
        if spec.type == "combo" or spec.type == 'combo-edit':  # can't get a good implementation of combo editing in ipython currently
            w = widgets.Dropdown(
                options=spec.options or [],
                value=spec.value,
                description=spec.label or "",
                layout=widgets.Layout(width="200px")
            )
            w.observe(lambda change, cb=spec.callback: self._callback(cb, change["new"]), names="value")

        elif spec.type == "checkbox":
            w = widgets.Checkbox(
                value=bool(spec.value),
                description=spec.label or ""
            )
            w.observe(lambda change, cb=spec.callback: self._callback(cb, change["new"]), names="value")

        elif spec.type == "range_slider":
            lo, hi = spec.range or (0.0, 1.0)
            w = widgets.FloatRangeSlider(
                value=tuple(spec.value),
                min=lo, max=hi,
                step=None,
                description=spec.label or "",
                layout=widgets.Layout(width="400px")
            )
            w.observe(lambda change, cb=spec.callback: self._callback(cb,change["new"]), names="value")

        elif spec.type == "slider":
            lo, hi = spec.range or (0.0, 1.0)
            w = widgets.FloatSlider(
                value=spec.value,
                min=lo, max=hi,
                step=None,
                description=spec.label or "",
                layout=widgets.Layout(width="400px")
            )
            w.observe(lambda change, cb=spec.callback: self._callback(cb, change["new"]), names="value")

        elif spec.type == "button":
            w = widgets.Button(description=spec.label or "")
            w.on_click(lambda btn, cb=spec.callback: self._callback(cb, None))

        else:
            w = widgets.HTML(f"<b>Unknown control {spec.name}</b>")

        return w


    def update_widget(self, spec, widget):
        if isinstance(spec, ControlSpec):
            if spec.type in {"combo", "combo-edit"}:
                widget.value = spec.value
            elif spec.type == "checkbox":
                widget.value = bool(spec.value)
            elif spec.type == "range_slider":
                lo, hi = spec.range or (0.0, 1.0)
                self.safe_update_slider_range(widget, lo, hi)
                wlo, whi = spec.value

                # seemingly need to set this after the range update has gone through, otherwise get
                # nonsense results in some cases
                self.call_later(JUPYTER_UI_LAG/2, lambda: setattr(widget, "value", (wlo, whi)))

            elif spec.type == "slider":
                lo, hi = spec.range or (0.0, 1.0)
                self.safe_update_slider_range(widget, lo, hi)

                # seemingly need to set this after the range update has gone through, otherwise get
                # nonsense results in some cases
                self.call_later(JUPYTER_UI_LAG/2, lambda: setattr(widget, "value", spec.value))
        else:
            for child_spec, child_widget in zip(spec.children, widget.children):
                self.update_widget(child_spec, child_widget)

    @classmethod
    def safe_update_slider_range(cls, slider, min_, max_):
        # sliders in ipywidgets seem to offer no option to update the range atomically. If one naively sets
        # min and max, the intermediate state can be invalid and raise an exception. Therefore, one needs
        # to first set a bounding range, then narrow back down to min, max.
        if slider.min == min_ and slider.max == max_:
            return
        bounding_min = min(min_, slider.min)
        bounding_max = max(max_, slider.max)
        slider.min = bounding_min
        slider.max = bounding_max
        slider.min = min_
        slider.max = max_




from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Tuple, Union
import matplotlib as mpl
import abc

from .. import config

@dataclass
class ControlSpec:
    name: str
    type: str  # 'combo' | 'combo-edit' | 'checkbox' | 'range_slider' | 'button'
    label: Optional[str] = None
    options: Optional[List[str]] = None
    value: Any = None
    range: Optional[Tuple[float, float]] = None
    callback: Callable[[Any], None] = lambda _: None

@dataclass
class LayoutSpec:
    type: str  # 'vbox' | 'hbox'
    children: List[Union['LayoutSpec', ControlSpec]]

class GenericController(abc.ABC):
    def __init__(self, visualizer, refresh_ui_callback: Optional[Callable] = None):
        self.visualizer = visualizer
        self._refresh_ui_callback = refresh_ui_callback

    @abc.abstractmethod
    def get_layout(self) -> LayoutSpec:
        pass

    def refresh_ui(self) -> None:
        if self._refresh_ui_callback is not None:
            self._refresh_ui_callback()

class ColorMapController(GenericController):
    """Controller description for standard color maps"""

    default_quantity_name = config.PROJECTED_DENSITY_NAME

    def __init__(self, visualizer, refresh_ui_callback: Optional[Callable] = None):
        super().__init__(visualizer, refresh_ui_callback)

    def get_colormap_list(self) -> List[str]:
        return list(mpl.colormaps.keys())

    def get_quantity_list(self) -> List[str]:
        names = sorted(self.visualizer.data_loader.get_quantity_names(), key=str.lower)
        return [self.default_quantity_name] + names

    def apply_auto(self) -> None:
        self.visualizer.colormap.autorange_vmin_vmax()
        self.refresh_ui()

    def apply_colormap(self, name: str) -> None:
        self.visualizer.colormap_name = name

    def apply_log_scale(self, state: bool) -> None:
        self.visualizer.log_scale = state
        self.visualizer.vmin, self.visualizer.vmax = self.visualizer.colormap.get_ui_range()

    def apply_quantity(self, name: str) -> None:
        new = None if name == self.default_quantity_name else name
        self.visualizer.quantity_name = new
        # other elements of the UI may need to be updated
        self.refresh_ui()

    def apply_slider(self, vmin: float, vmax: float) -> None:
        self.visualizer.vmin, self.visualizer.vmax = vmin, vmax

    def get_layout(self) -> LayoutSpec:
        cmap = self.visualizer.colormap_name
        qty = self.visualizer.quantity_name or self.default_quantity_name
        ui_range = self.visualizer.colormap.get_ui_range()

        return LayoutSpec(
            type="vbox",
            children=[
                LayoutSpec("hbox", [
                    ControlSpec("colormap", "combo", options=self.get_colormap_list(),
                                value=cmap, callback=self.apply_colormap),
                    ControlSpec("quantity", "combo-edit", options=self.get_quantity_list(),
                                value=qty, callback=self.apply_quantity),
                    ControlSpec("log", "checkbox", label="Log scale",
                                value=self.visualizer.log_scale, callback=self.apply_log_scale),
                ]),
                LayoutSpec("hbox", [
                    ControlSpec("range", "range_slider", value=(self.visualizer.vmin, self.visualizer.vmax),
                                range=ui_range, callback=lambda vv: self.apply_slider(*vv)),
                    ControlSpec("auto", "button", label="Auto",
                                callback=lambda _: self.apply_auto()),
                ]),
            ],
        )

class RGBMapController(GenericController):
    """Controller description for RGB (stellar rendering) outputs"""

    def __init__(self, visualizer, refresh_ui_callback: Optional[Callable] = None):
        super().__init__(visualizer, refresh_ui_callback)

    def get_state(self) -> dict:
        cmap = self.visualizer.colormap
        return {
            "mag_range": (cmap.min_mag, cmap.max_mag),
            "gamma": cmap.gamma,
        }

    def apply_mag_range(self, mag_pair: Tuple[float, float]) -> None:
        lo, hi = mag_pair
        cmap = self.visualizer.colormap
        cmap.min_mag, cmap.max_mag = lo, hi

    def apply_gamma(self, g: float) -> None:
        self.visualizer.colormap.gamma = g

    def get_layout(self) -> LayoutSpec:
        st = self.get_state()
        return LayoutSpec(
            type="vbox",
            children=[
                ControlSpec(
                    name="mag_range",
                    type="range_slider",
                    label='mag/"^2',
                    range=(15.0,40.0),
                    value=st["mag_range"],
                    callback=lambda v: self.apply_mag_range(v),
                ),
                ControlSpec(
                    name="gamma",
                    type="slider",
                    label="gamma",
                    range=(0.25, 8.0),          # fixed slider‐domain
                    value=st["gamma"],          # initial value = current state
                    callback=lambda v: self.apply_gamma(v),
                ),
            ],
        )
    


"""Classes to keep track of the cellular layout of a simulation"""

import numpy as np
import pynbody

from pynbody.filt import geometry_selection

class CellLayout:
    """Class to keep track of segmentation of a simulation into cells"""
    def __init__(self, centres: np.ndarray, offsets: np.ndarray, lengths: np.ndarray):
        self._centres = np.ascontiguousarray(centres)
        self._offsets = offsets
        self._lengths = lengths
        self._num_particles = lengths.sum()
        self._cell_size = np.linalg.norm(self._centres[1]-self._centres[0])

    def randomize_within_cells(self):
        """Get a reordering of the particles which randomizes the order within cells, but leaves the cell structure"""
        total_len = self._lengths.sum()
        reordering = np.empty(total_len, dtype=np.uintp)
        for offset, length in zip(self._offsets, self._lengths):
            # randomize the order of the particles within this cell
            reordering[offset:offset+length] = np.random.permutation(length) + offset
        return reordering

    def cells_in_sphere(self, centre: tuple[float, float, float], radius: float) -> np.ndarray:
        """Get the indices of the cells that are within a sphere of given centre and radius"""
        expand_radius = self._cell_size*np.sqrt(3.0)
        offsets = self._centres - centre
        selection = np.linalg.norm(offsets, axis=1) < (radius + expand_radius)
        return np.where(selection)[0]

    def cell_index_from_offset(self, offset: int) -> int:
        """Get the cell index from the offset of a particle"""

        cell_index = np.searchsorted(self._offsets, offset, side='right') - 1
        if cell_index < 0 or cell_index >= len(self._lengths):
            raise ValueError("Offset is out of bounds")
        return cell_index

    def cell_slice(self, cell_index: int) -> slice:
        """Get the indices of the particles in a given cell"""
        start = self._offsets[cell_index]
        end = start + self._lengths[cell_index]
        return slice(start, end)

    def get_num_cells(self):
        """Get the total number of cells"""
        return len(self._lengths)

    def get_num_particles(self):
        """Get the total number of particles"""
        return self._num_particles

    def get_cell_length(self, cell_index: int | np.ndarray[int]) -> int | np.ndarray[int]:
        """Get the length of a given cell"""
        return self._lengths[cell_index]

    def get_cell_offset(self, cell_index: int) -> int:
        """Get the offset of a given cell"""
        return self._offsets[cell_index]

    @classmethod
    def from_positions(cls, particle_positions: np.ndarray, box_min: float, box_max: float, nside: int):
        """Create a CellLayout object from the positions of the particles with arbitrary ordering

        Parameters
        ----------

        particle_positions: array of the positions of the particles (Nx3)
        box_min: minimum coordinate of the box (for all 3 dimensions)
        box_max: maximum coordinate of the box (for all 3 dimensions)
        nside: number of cells in each of the 3 dimensions, e.g. nside=10 implies 10^3 cells in total

        Returns
        -------

        cell_layout, particle_ordering:
            cell_layout: CellLayout object
            particle_ordering: array of the ordering of the particles to put them into the cells
        """

        if particle_positions.min()<box_min or particle_positions.max()>=box_max:
            raise ValueError("Particle positions are outside the box")

        # get the cell size
        cell_size = (box_max - box_min) / nside

        # get the centre of the first cell
        cell_cen0 = box_min + cell_size / 2

        # get the cell centres
        centres = np.mgrid[cell_cen0:box_max:cell_size,
                           cell_cen0:box_max:cell_size,
                           cell_cen0:box_max:cell_size].reshape(3, -1).T

        # figure out the cell x,y,z indices of each particle
        pos_indices = np.floor((particle_positions - box_min) / cell_size).astype(np.intp)

        if pos_indices.min() < 0 or pos_indices.max() >= nside:
            raise ValueError("Particle positions are too close to edge of box; expand box size")

        # get the cell index of each particle
        cell_indices = pos_indices[:, 2] + nside * (pos_indices[:, 1] + nside * pos_indices[:, 0])
        # sort the particles by cell index
        ordering = np.argsort(cell_indices)

        # figure out the segmentation
        lengths = np.bincount(cell_indices, minlength=nside ** 3)
        assert len(lengths) == len(centres), "Logic error within from_positions"

        offsets = np.cumsum(lengths) - lengths

        return cls(centres, offsets, lengths), ordering